import base64
import json
import zlib
from datetime import datetime

import pytz
from apscheduler.triggers.cron import CronTrigger

from bspider.core.api import BaseService, GetSuccess, NotFound, AgentMixIn

from .impl.tools_impl import ToolsImpl


class ToolsService(BaseService, AgentMixIn):

    def __init__(self):
        self.impl = ToolsImpl()
        self.tz = pytz.timezone(self.frame_settings['TIMEZONE'])

    def get_node_list(self):
        """获取节点列表"""
        infos = self.impl.get_node_list()
        return GetSuccess(msg='get code success', data=infos)

    def get_code_list(self, code_type):
        infos = self.impl.get_code_list(code_type)
        return GetSuccess(msg='get code success', data=infos)

    def validate(self, valid_type, data):
        if valid_type == 'crontab':
            try:
                cron = CronTrigger.from_crontab(data)
                cur_time = None
                now = datetime.now(self.tz)
                result = list()
                for i in range(5):
                    cur_time = cron.get_next_fire_time(cur_time, now)
                    now = cur_time
                    result.append(cur_time.strftime('%Y-%m-%d %H:%M:%S'))
                return GetSuccess(msg='validate complete', data={'valid': True, 'time': result})
            except Exception:
                return GetSuccess(msg='validate complete', data={'valid': False})
        return NotFound(msg='unknow validate type', errno=60001)

    def node_status(self, ip):
        """调用接口返回节点状态，兼职探活"""
        return GetSuccess(data=self.op_get_node_status(ip))

    def get_parser_exception(self, project_id):
        infos = self.impl.get_parser_exception(project_id)
        for info in infos:
            self.datetime_to_str(info)
            if info['exception']:
                info['exception'] = info['exception'].replace('\n', '<br>')
        return GetSuccess(data=infos)

    def get_downloader_exception(self, project_id):
        infos = self.impl.get_downloader_exception(project_id)
        for info in infos:
            self.datetime_to_str(info)
            if info['exception']:
                info['exception'] = info['exception'].replace('\n', '<br>')
        return GetSuccess(data=infos)

    def get_crawl_detail(self, tag, source, data):
        if tag == 'sign':
            infos = self.impl.get_crawl_detail(data, source)
        else:
            sign = self.impl.get_sign_by_url(data)
            if sign:
                infos = self.impl.get_crawl_detail(sign, source)
            else:
                return NotFound(errno=60002, msg=f'NotFound url:{data}')

        for info in infos:
            self.datetime_to_str(info)
            if info['response'] is not None:
                info['response'] = json.loads(zlib.decompress(base64.b64decode(infos[0]['response'])).decode())
            return GetSuccess(data=info)
        return NotFound(errno=60002, msg=f'NotFound tag:{data}')

    def get_node_detail(self):
        return GetSuccess(data=self.impl.get_node_detail()[0])

    def get_exception_project(self):
        return GetSuccess(data=self.impl.get_exception_project())

    def get_current_user(self, user_id):
        infos = self.impl.get_user_by_id(user_id)

        for info in infos:
            self.datetime_to_str(info)

        if len(infos):
            return GetSuccess(msg='get user success', data=infos[0])
        return NotFound(errno=10006, msg='invalid user')




    def get_request_track(self, url=None, sign=None):
        if url:
            sign = self.impl.get_sign_by_url(url)

        if sign:
            track_info = self.impl.get_reqest_track(sign)
            return 0, f'查询成功', track_info

    def parser_error(self, error):
        for err in error:
            err['size'] = 'large'
            err['timestamp'] = err['crawl_time']
            if err['status'] == 599:
                err['type'] = 'warning'
            else:
                err['type'] = 'danger'

            if err['status'] == -1:
                err['status'] = 'FAIL'
            if err['exception']:
                err['exception'] = err['exception'].replace('\n\n', '<br>')
        return error



from .base_pipeline import BasePipeline
from .base_extractor import BaseExtractor
from .work import run_parser
from .item import *

__all__ = [
    'BasePipeline',
    'BaseExtractor',
    'run_parser',
    'Item',
    'MySQLSaverItem',
    'RabbitMQSaverItem',
    'RedisSaverItem'
]


"""
@description=使用 bproxypool服务 为请求头添加proxy
cddcsdcddvffdvv
"""
import json

import aiohttp.client_exceptions
from bspider.downloader import BaseMiddleware
from bspider.utils.exceptions import DownloaderError, MiddleWareParamError


class BProxyPoolMiddleware(BaseMiddleware):
    """
    使用 bproxypool服务 为请求头添加proxy
    """

    def __init__(self, project, settings, log):
        super().__init__(project, settings, log)
        bproxypool_url = self.settings.get('api')
        if bproxypool_url is None:
            raise MiddleWareParamError('\'api\' is Required like http://127.0.0.1:80/proxy/')
        self.timeout = self.settings.get('timeout', 3)
        self.retry_time = self.settings.get('retry_time', 3)
        virtual_pool = self.settings.get('virtual_proxy_pool')
        self.cool_down_time = self.settings.get('proxy_cool_down_time')
        ignore = self.settings.get('ignore_cool_down_http_code', list())
        ignore.append(599)
        self.ignore = set(ignore)
        if virtual_pool:
            self.proxy_url = f'{bproxypool_url}{virtual_pool}'
        else:
            self.proxy_url = bproxypool_url

    # by coroutine
    # async process_request(self, request):
    async def process_request(self, request):
        """
        执行下载前的操作
        返回Response类型，作为response完成下载，跳过后面的下载阶段，直接进入下载后的中间件
        返回空或其他返回值 继续下一个中间件
        """
        for index in range(self.retry_time):
            try:
                proxy_req = await self.get_proxy()
                request.proxy = {
                    'proxy': 'http://{}'.format(proxy_req['data']['proxy']),
                    'source': proxy_req['data']['source']
                }
                self.log.info('success add proxy in {} {}'.format(
                    request,
                    request.proxy
                ))
                return request
            except Exception as e:
                self.log.warning(f'get proxy failed: {e} {request} retry:{index}')

        raise DownloaderError('Proxy get failed, please check proxy service!')

    async def process_exception(self, request, e, response):
        if isinstance(e, aiohttp.ClientProxyConnectionError):
            await self.delete_proxy(request.proxy)
            self.log.info('{proxy} {source} is invalid delete it'.format(
                **request.proxy
            ))
        elif response.status not in self.ignore:
            await self.cool_down_proxy(request.proxy)
            self.log.info('{proxy} {source} success to cool down'.format(
                **request.proxy
            ))
        return response

    async def get_proxy(self) -> dict:
        temp_timeout = aiohttp.ClientTimeout(total=self.timeout)
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method='GET',
                    url=self.proxy_url,
                    timeout=temp_timeout
            ) as resp:
                # 挂起等待下载结果
                return json.loads(await resp.text())

    async def delete_proxy(self, proxy: dict) -> bool:
        temp_timeout = aiohttp.ClientTimeout(total=self.timeout)
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method='DELETE',
                    url=self.proxy_url,
                    timeout=temp_timeout,
                    data=proxy
            ) as resp:
                # 挂起等待下载结果
                return resp.status == 204

    async def cool_down_proxy(self, proxy: dict) -> bool:
        temp_timeout = aiohttp.ClientTimeout(total=self.timeout)
        data = {'proxy': proxy['proxy']}
        if self.cool_down_time:
            data['expire'] = self.cool_down_time
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method='PATCH',
                    url=self.proxy_url,
                    timeout=temp_timeout,
                    data=data
            ) as resp:
                # 挂起等待下载结果
                return resp.status == 200



from bspider.utils.logger import LoggerPool
from bspider.utils.system import System

log = LoggerPool().get_logger(key='agent', fn='agent', module=f'agent', ip=System.ip_msg)
# 注册一个Auth模块的句柄，用来实现复用模块日志输出至不同文件的功能
LoggerPool().get_logger(key='api_auth', fn='agent', module='api_auth')



from flask import Blueprint

from bspider.core.api import auth
from bspider.utils.tools import change_dict_key
from .validators import PageForm
from bspider.master.service.cron import CronService
from .validators.cron_forms import AddForm, UpdateForm

cron = Blueprint('cron_bp', __name__)

cron_service = CronService()


@cron.route('/cron', methods=['POST'])
@auth.login_required
def add():
    form = AddForm()
    return cron_service.add_cron(**change_dict_key('type', 'cron_type', form.to_dict()))


@cron.route('/cron/<int:cron_id>', methods=['DELETE'])
@auth.login_required
def delete(cron_id):
    return cron_service.delete_job(cron_id)


@cron.route('/cron/<int:cron_id>', methods=['PATCH'])
@auth.login_required
def update(cron_id):
    form = UpdateForm()
    return cron_service.update_job(cron_id, changes=form.to_dict())


@cron.route('/cron/<int:cron_id>', methods=['GET'])
@auth.login_required
def get(cron_id):
    return cron_service.get_job(cron_id)

@cron.route('/cron', methods=['GET'])
@auth.login_required
def gets():
    form = PageForm()
    return cron_service.get_jobs(**form.to_dict())



"""
常用工具封装
"""
import asyncio
import datetime
import hashlib
import json
import re
import time

from apscheduler.triggers.cron import CronTrigger
from apscheduler.util import datetime_to_utc_timestamp

from bspider.utils.exceptions import ModuleError


def make_sign(url, salt=''):
    return '{}-{}'.format(hashlib.md5(f'{url}{salt}'.encode('utf-8')).hexdigest(), time.time())


def coroutine_result(coroutine, loop=None):
    """a function to get coroutine return"""
    if loop is None:
        loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(coroutine, loop=loop)
    loop.run_until_complete(task)
    return task.result()


def change_dict_key(cur_key, replace_key, d: dict) -> dict:
    if cur_key in d:
        d[replace_key] = d.pop(cur_key)
    return d


def make_fields_values(data: dict) -> tuple:
    """
    给定字典，返回fields 和 values
    :param info: dict
    :return:
    """
    fields = list()
    values = list()
    for key, value in data.items():
        fields.append(' `%s`=%%s ' % (key))
        if isinstance(value, dict) or isinstance(value, list):
            values.append(json.dumps(value))
        else:
            values.append(value)

    return ','.join(fields), tuple(values)


def find_class_name_by_content(content):
    reg = re.compile('class (?P<class_name>.*?)\((?P<sub_class_name>.*?)\):').search(content)
    if reg:
        tmp = reg.groupdict()
        return tmp['class_name'], tmp['sub_class_name']
    raise ModuleError('content can\'t find class_name and sub_class_name -> \n%s' % (content[0: 100] + ' ...'))


def module_name2class_name(module_name):
    """
    python 的模块名改为类名
    demo_pipeline -> DemoPipeline
    """
    return ''.join([word.title() for word in module_name.lower().split('_')])

def class_name2module_name(class_name):
    """
    python 的模块名改为类名
    DemoPipeline -> demo_pipeline
    """
    pattern = re.compile(r'([A-Z]{1})')
    return re.sub(pattern, r'_\1', class_name).lower().replace('_', '', 1)

def get_crontab_next_run_time(crontab, tz = None):
    crontab = CronTrigger.from_crontab(crontab)
    now = datetime.datetime.now(tz)
    next_run_time = crontab.get_next_fire_time(None, now)
    return datetime_to_utc_timestamp(next_run_time), next_run_time




from bspider.core.custom_module import BaseCustomModule


class BasePipeline(BaseCustomModule):

    def process_item(self, item):
        """
        对item进行处理,各个pipeline重写
        注意按照scrapy的风格
        可以yield返回也可以return
        返回类型为自订定义的类型，会传递给后面的pipeline，返回None将会丢弃这个item
        :param item:
        :return:
        """
        return item



import asyncio
import json
import random
import sys
import traceback

from bspider.config.default_settings import EXCHANGE_NAME
from bspider.utils.rabbitMQ import AioRabbitMQClient
from bspider.utils.sign import Sign
from bspider.utils.tools import find_class_name_by_content

from .agent_cache import AgentCache
from .project import Project


class BaseMonitor(object):
    exchange = ''

    def __init__(self, log, log_fn, mq_client: AioRabbitMQClient):
        self.log = log
        self.__cache = AgentCache()
        self.log.info(f'init cache success! {self.__cache}')
        self.mq_client = mq_client
        self.projects = dict()
        self.__weight = None
        self.__total_sum = 0
        self.log_fn = log_fn

    async def sync_config(self):
        """从cache 同步任务数据"""
        while True:
            try:
                await self.__sync_config()
            except Exception as e:
                tp, msg, tb = sys.exc_info()
                e_msg = ''.join(traceback.format_exception(tp, msg, tb))
                self.log.error(f'sync project failed:{e_msg}')
                self.log.error(f'sync project failed:{e}')
            await asyncio.sleep(2)

    async def __sync_config(self):
        """
        如果不是第一次同步
        获取当前时间戳 -10s
        :return:
        """
        tmp_projects = dict()
        tmp_weight = dict()
        projects = self.__cache.get_projects()
        total_sum = 0

        while len(projects):
            info = projects.pop()
            if await self.mq_client.get_queue_message_count(queue='{}_{}'.format(self.exchange, info['id'])) != 0:
                tmp_weight[info['id']] = info['rate']
                total_sum += info['rate']

            worker_obj = self.projects.get(info['id'])
            if isinstance(info['config'], str):
                info['config'] = json.loads(info['config'])
            project = Project(
                settings=info['config'],
                pipeline_serializer_method=self.code_id_to_content,
                middleware_serializer_method=self.code_id_to_content
            )
            project.project_id = info['id']

            # code_id映射为中间件代码
            if self.exchange == EXCHANGE_NAME[2]:
                sign = Sign(project_timestamp=info['timestamp'], module=str(project.parser_settings.pipeline))
            else:
                sign = Sign(project_timestamp=info['timestamp'], module=str(project.downloader_settings.middleware))

            if worker_obj is None or worker_obj.sign != sign:
                tmp_projects[project.project_id] = self.get_work_obj(project, sign=sign)
            else:
                tmp_projects[info['id']] = worker_obj

        self.projects = tmp_projects
        self.log.debug('sync projects success {}'.format(len(self.projects)))
        self.__weight = sorted(tmp_weight.items(), key=lambda d: d[1], reverse=False)
        self.__total_sum = total_sum
        self.log.debug('sync weight success {} total num {}'.format(len(self.__weight), self.__total_sum))

    def choice_project(self) -> int:
        """
        # 根据project的权重值随机选取一个
        :return:选取的值在原列表里的索引
        """
        if self.__weight is not None and self.__total_sum:
            seed = random.randint(0, self.__total_sum - 1)
            for project_id, weight in self.__weight:
                seed -= weight
                if seed < 0:
                    return project_id

    def code_id_to_content(self, code_id):
        content = self.__cache.get_code(code_id)[0]['content']
        cls_name, _ = find_class_name_by_content(content)
        return cls_name, content

    def get_work_obj(self, project: Project, sign: Sign):
        """继承重写 -> 根据配置返回对象"""
        pass



"""
    封装对cache中project信息的操作为api
"""
from flask import Blueprint

from bspider.core.api import auth
from .validators.project_form import AddForm, UpdateForm
from bspider.agent.service.project import ProjectService

project = Blueprint('project_bp', __name__)
project_service = ProjectService()


@project.route('/project', methods=['POST'])
@auth.login_required
def add_project():
    form = AddForm()
    return project_service.add_project(**form.to_dict())

@project.route('/project/<int:project_id>', methods=['PATCH'])
@auth.login_required
def update_project(project_id):
    form = UpdateForm()
    return project_service.update_project(project_id, form.to_dict())

@project.route('/project/<int:project_id>', methods=['DELETE'])
@auth.login_required
def delete_project(project_id):
    return project_service.delete_project(project_id)


@project.route('/project', methods=['GET'])
@auth.login_required
def get_projects():
    return project_service.get_projects()


@project.route('/project/<int:project_id>', methods=['GET'])
def get_project(project_id):
    return project_service.get_project(project_id)



import asyncio
import json
import traceback

import sys

from bspider.core import BaseManager
from bspider.config.default_settings import EXCHANGE_NAME
from bspider.http import Request, ERROR_RESPONSE

from .downloader_monitor import DownloaderMonitor


def run_downloader(unique_sign, coro_num=1):
    """A factory to make a download process"""
    dm = DownloaderManager(unique_sign, DownloaderMonitor, coro_num)
    dm.run()


class DownloaderManager(BaseManager):
    """全异步操作的下载器管理器"""
    manager_type = 'downloader'
    exchange = EXCHANGE_NAME[1]

    async def do_work(self):
        try:
            while True:
                downloader = self.monitor.projects.get(self.monitor.choice_project())
                if downloader is None:
                    # 防止协程抢占无法轮换
                    await asyncio.sleep(1)
                    continue

                # 异常占位符
                e_msg = None
                response = ERROR_RESPONSE
                sign = False
                async with self.broker.mq_client.session() as session:
                    msg_id, data = await session.recv_msg(f'{self.exchange}_{downloader.project_id}')
                    if msg_id:
                        try:
                            request = Request.loads(json.loads(data))
                            self.log.info(f'success get a new Request: {request}')
                            if request.life_cycle is None or request.life_cycle > 0:
                                response, sign, e_msg = await downloader.download(request)
                                if isinstance(request.life_cycle, int):
                                    response.request.life_cycle -= 1
                            else:
                                self.log.warning(f'Request life_cycle <= 0 ignore it: {request}')
                        except Exception:
                            tp, msg, tb = sys.exc_info()
                            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
                        session.ack(msg_id)

                # 分开写是因为要提前释放channel,防止channel 被用光影响性能
                if msg_id:
                    await self.broker.set_response(response, downloader.project_id)
                    if e_msg or not sign:
                        self.log.info(
                            f'project:project_id->{downloader.project_id} project_name->{downloader.project_name} fail download: {request}')
                        await self._save_error_result(response, downloader.project_name, downloader.project_id, e_msg)
                    else:
                        # 持久化下载结果
                        await self._save_success_result(response, downloader.project_name, downloader.project_id)
                        self.log.info('project:project_id->{} project_name->{} complete download: {}'.format(
                            downloader.project_id, downloader.project_name, response.url))
        except Exception:
            tp, msg, tb = sys.exc_info()
            e_msg = ''.join(traceback.format_exception(tp, msg, tb))
            self.log.error(f'coro error:{e_msg}')


if __name__ == '__main__':
    run_downloader('downloader01')



from datetime import datetime

import pytz

from bspider.config import FrameSettings


class BaseService(object):
    frame_settings = FrameSettings()
    tz = pytz.timezone(frame_settings['TIMEZONE'])

    def datetime_to_str(self, d: dict) -> None:
        for key, value in d.items():
            if isinstance(value, datetime):
                d[key] = value.strftime("%Y-%m-%d %H:%M:%S")
            if isinstance(value, dict):
                self.datetime_to_str(value)



from wtforms import StringField, IntegerField

from bspider.core.api import BaseForm, ParamRequired


class AddForm(BaseForm):
    project_id = IntegerField(validators=[ParamRequired()])
    code_id = IntegerField(validators=[ParamRequired()])
    type = StringField(validators=[ParamRequired()])
    trigger = StringField(validators=[ParamRequired()])
    description = StringField(validators=[ParamRequired()])


class UpdateForm(BaseForm):
    project_id = IntegerField()
    code_id = IntegerField()
    type = StringField()
    trigger = StringField()
    description = StringField()



from pymysql import IntegrityError
from werkzeug.security import generate_password_hash, check_password_hash

from bspider.core.api import BaseService, ParameterException, NotFound, Conflict, GetSuccess, PostSuccess, DeleteSuccess, PatchSuccess
from bspider.core.api.auth.token import make_token

from bspider.master.server import log
from bspider.master.service.impl.user_impl import UserImpl


class UserService(BaseService):

    def __init__(self):
        self.impl = UserImpl()

    def login_by_identity(self, identity, password):
        infos = self.impl.get_user(identity)

        for info in infos:
            self.datetime_to_str(info)

        if len(infos) == 1:
            info = infos[0]
            if info['status'] == 0:
                return Conflict(errno=10009, msg='this identity was closed, please call admin!')

            if check_password_hash(info.pop('password'), password):
                info['token'] = make_token(info['id'], info['role'])
                log.info(f'user->identity:{identity} login success')
                return GetSuccess(msg='login success', data=info)
            else:
                log.info(f'user->identity:{identity} login failed:invalid password')
                return ParameterException(errno=10005, msg='invalid password')
        elif len(infos) == 0:
            log.info(f'user->identity:{identity} login failed:invalid user')
            return NotFound(errno=10006, msg='invalid user')
        else:
            return Conflict(errno=10008, msg='query exceptions')

    def add_user(self, identity, username, password, role, email, phone, status):
        data = {
            'identity': identity,
            'username': username,
            'password': generate_password_hash(password),
            'role': role,
            'email': email,
            'phone': phone,
            'status': status
        }
        try:
            user_id = self.impl.add_user(data)
            data['user_id'] = user_id
            data.pop('password')
            log.info(f'user->identity:{identity} add success')
            return PostSuccess(msg=f'add user {username} success', data=data)
        except IntegrityError as e:
            if e.args[0] == 1062:
                log.info(f'user->identity:{identity} add failed:user is already exist')
                return Conflict(errno=10007, msg='user is already exist')
            raise e

    def remove_user(self, user_id):
        self.impl.remove_user(user_id)
        log.warning(f'user->user_id:{user_id} delete success')
        return DeleteSuccess()

    def update_user(self, user_id, **kwargs):
        if 'password' in kwargs:
            kwargs['password'] = generate_password_hash(kwargs['password'])
        self.impl.update_user(user_id, kwargs)
        log.info(f'user->user_id:{user_id}->{kwargs} update success')
        return PatchSuccess(msg='update user success!')

    def get_users(self, page, limit, search, sort):
        if sort.upper() not in ['ASC', 'DESC']:
            return ParameterException(msg='sort must `asc` or `desc`')

        infos, total = self.impl.get_users(page, limit, search, sort)

        for info in infos:
            info.pop('password')
            self.datetime_to_str(info)

        return GetSuccess(
            msg='get user list success!',
            data={
                'items': infos,
                'total': total,
                'page': page,
                'limit': limit
            })

    def get_user(self, user_id):
        infos = self.impl.get_user_by_id(user_id)

        for info in infos:
            self.datetime_to_str(info)

        if len(infos):
            return GetSuccess(msg='get user success', data=infos[0])
        return NotFound(errno=10006, msg='invalid user')



class BaseHttp(object):

    def _set_url(self, url):
        if isinstance(url, str):
            return url
        else:
            raise TypeError('%s url must be str, got: %s' % (type(self).__name__, type(url).__name__))

    def _set_headers(self, headers):
        if headers is None:
            return dict()
        elif not isinstance(headers, dict):
            raise TypeError("%s headers must be dict. " % (type(self).__name__))
        else:
            return headers

    def _set_cookies(self, cookies):
        if cookies is None:
            return
        elif not isinstance(cookies, dict):
            raise TypeError("%s cookies must be dict or None. " % (type(self).__name__))
        else:
            return cookies

    def _set_method(self, method):
        if method is None:
            return 'GET'
        if isinstance(method, str):
            method = method.upper()
            if method in 'GET, POST, PUT ,PATCH, OPTIONS, HEAD, DELETE':
                return method
            else:
                raise TypeError("%s method: %s is not support. " % (type(self).__name__, method))

        raise TypeError("%s method must be str. " % (type(self).__name__))

    def _set_meta(self, meta):
        if meta is None:
            return dict()
        elif not isinstance(meta, dict):
            raise TypeError("%s meta must be dict or None. " % (type(self).__name__))
        else:
            return meta

    def _set_data(self, data):
        if data is None:
            return
        elif not isinstance(data, dict):
            raise TypeError("%s data must be dict or None. " % (type(self).__name__))
        else:
            return data

    def _set_params(self, params):
        if params is None:
            return
        elif not isinstance(params, dict):
            raise TypeError("%s params must be dict or None. " % (type(self).__name__))
        else:
            return params

    # Response API
    def _set_text(self, text):
        if text is None:
            return ''
        elif not isinstance(text, str):
            raise TypeError("%s text must be str. " % (type(self).__name__))
        else:
            return text

    def _set_callback(self, callback):
        if callable(callback):
            return callback.__name__
        elif isinstance(callback, str):
            return callback
        else:
            raise TypeError('callback must be callable or str')

    def _set_errback(self, errback):
        if callable(errback):
            return errback.__name__
        elif isinstance(errback, str) or errback is None:
            return errback
        else:
            raise TypeError('error callback must be callable or str')

    def _set_proxy(self, proxy):
        if proxy is None:
            return
        elif not isinstance(proxy, dict):
            raise TypeError("%s proxy must be dict or None. " % (type(self).__name__))
        elif 'proxy' not in proxy:
            raise KeyError("%s proxy must have key:\'proxy\'. " % (type(self).__name__))
        else:
            return proxy



from wtforms import IntegerField, StringField

from bspider.core.api import BaseForm, ParamRequired


class AddForm(BaseForm):
    code_id = IntegerField(validators=[ParamRequired()])
    content = StringField(validators=[ParamRequired()])


class UpdateForm(BaseForm):
    content = StringField()
    project = StringField(default='')

    def validate_project(self, value):
        value.data = value.data.split(',')



import sys
from types import GeneratorType

from bspider.http import Response
from bspider.parser import BasePipeline
from bspider.utils.exceptions import ExtractorCallbackError


class BaseExtractor(BasePipeline):

    def start_url(self) -> list:
        """生成任务起始的URL"""
        raise NotImplementedError

    def process_item(self, response):
        if isinstance(response, Response):
            if response.callback:
                try:
                    items = getattr(self, response.callback)(response)
                    if not isinstance(items, GeneratorType):
                        return items

                    for item in items:
                        yield item
                except Exception as e:
                    tp, msg, tb = sys.exc_info()
                    e.with_traceback(tb)
                    yield getattr(self, response.errback)(response, e, msg)
            else:
                raise ExtractorCallbackError('Cannot find callback %s in extractor %s' %
                                             (self.__class__.__name__, response.callback))
        yield response

    def parser(self, response: Response):
        """
        此方法需要被重构
        当一个 response 对象没有指定 callback 方法时此方法将被执行
        """
        pass

    def errback(self, response: Response, e: Exception, traceback: str):
        raise e



from wtforms import StringField, IntegerField
from wtforms.validators import length

from bspider.core.api import BaseForm, ParamRequired
from .enums import ClientTypeEnum


class LoginForm(BaseForm):
    identity = StringField(validators=[ParamRequired(), length(max=32, min=5)])
    password = StringField()
    client_type = IntegerField(default=101)

    def validate_client_type(self, value):
        try:
            value.data = ClientTypeEnum(value.data).name
        except ValueError as e:
            raise e


class RegisterForm(BaseForm):
    identity = StringField(validators=[ParamRequired(), length(max=32, min=5)])
    username = StringField(validators=[ParamRequired(), length(max=32, min=1)])
    password = StringField(validators=[length(max=32, min=5)])
    role = StringField(validators=[ParamRequired()])
    email = StringField()
    phone = StringField()
    status = IntegerField(default=1)


class UpdateForm(BaseForm):
    username = StringField()
    password = StringField()
    role = StringField()
    email = StringField()
    phone = StringField()
    status = IntegerField()



from bspider.core.api import BaseImpl
from bspider.master import log


class UserImpl(BaseImpl):

    def get_user(self, identity: str):
        sql = f'select `id`, `identity`, `username`, `password`, `role`, `email`, `phone`, `status`, `create_time`, `update_time` ' \
              f'from {self.user_table} where `identity`=%s;'
        return self.mysql_client.select(sql, identity)

    def get_user_by_id(self, user_id: int):
        sql = f'select `id`, `identity`, `username`, `role`, `email`, `phone`, `status`, `create_time`, `update_time` ' \
              f'from {self.user_table} where `id`=%s;'
        return self.mysql_client.select(sql, user_id)

    def add_user(self, data):
        fields, values = self.make_fv(data)
        sql = f'insert into {self.user_table} set {fields};'
        return self.mysql_client.insert(sql, values, lastrowid=True)

    def remove_user(self, user_id):
        sql = f"update {self.user_table} set `status`=-1 where `id` = '{user_id}';"
        return self.mysql_client.update(sql)

    def update_user(self, user_id, data):
        fields, values = self.make_fv(data)
        sql = f"update {self.user_table} set {fields} where `id` = '{user_id}';"
        return self.mysql_client.update(sql, values)

    def get_users(self, page, limit, search, sort):
        start = (page - 1) * limit
        fields = self.make_search(search)
        if len(fields):
            sql = f'select `id`, `identity`, `username`, `password`, `role`, `email`, `phone`, `status`, `create_time`, `update_time` ' \
                  f'from {self.user_table} where {fields} order by `id` {sort} limit {start},{limit};'
        else:
            sql = f'select `id`, `identity`, `username`, `password`, `role`, `email`, `phone`, `status`, `create_time`, `update_time` ' \
                  f'from {self.user_table} order by `id` {sort} limit {start},{limit};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql), self.total_num(search, self.user_table)



"""
定时任务子模块
# 会初始化一个定时任务句柄
提供对定时任务的基本操作
-- 这里采用额外开启一个线程来控制整个定时任务模块
-- 考虑到多核cpu和一些 cpu密集型程序，
-- 这里采用 多线程的方式执行定时任务
"""
from datetime import datetime

from apscheduler.triggers.cron import CronTrigger
from apscheduler.util import datetime_to_utc_timestamp, obj_to_ref, utc_timestamp_to_datetime
from pymysql import IntegrityError

from bspider.core.api import BaseService, Conflict, PostSuccess, PatchSuccess, DeleteSuccess, GetSuccess, NotFound, \
    ParameterException
from bspider.bcron.todo import do
from bspider.master import log
from bspider.master.service.impl.cron_impl import CronImpl
from bspider.utils.tools import get_crontab_next_run_time


class CronService(BaseService):

    def __init__(self):
        self.impl = CronImpl()

    def add_cron(self, project_id, code_id, cron_type, trigger, description):
        timestamp, next_run_time = get_crontab_next_run_time(trigger, self.tz)
        value = {
            'project_id': project_id,
            'code_id': code_id,
            'type': cron_type,
            'trigger': trigger,
            'trigger_type': 'cron',
            'func': obj_to_ref(do),
            'executor': 'thread_pool',
            'description': description,
            'next_run_time': timestamp,
        }
        try:
            cron_id = self.impl.add_job(data=value)
            log.info(f'cron job->project_id:{project_id}-code_id:{code_id} add success')
            return PostSuccess(msg='add cron job success', data={'cron_id': cron_id})
        except IntegrityError:
            log.error(f'cron job->project_id:{project_id}-code_id:{code_id} is already exist')
            return Conflict(msg='cron job is already exist', errno=50001)

    def update_job(self, cron_id, changes):
        self.impl.update_job(cron_id, changes)
        log.info(f'update cron job->cron_id:{cron_id}->{changes} success')
        return PatchSuccess(msg=f'cron job update success')

    def delete_job(self, cron_id):
        self.impl.delete_job(cron_id)
        log.info(f'delete cron job->cron_id:{cron_id} success')
        return DeleteSuccess()

    def get_job(self, cron_id):
        infos = self.impl.get_job(cron_id)

        for info in infos:
            info['next_run_time'] = utc_timestamp_to_datetime(info['next_run_time']).astimezone(self.tz)
            self.datetime_to_str(info)

        if len(infos):
            return GetSuccess(msg='get cron job success', data=infos[0])
        return NotFound(msg='job is not exist', errno=50001)

    def get_jobs(self, page, limit, search, sort):
        if sort.upper() not in ['ASC', 'DESC']:
            return ParameterException(msg='sort must `asc` or `desc`')

        infos, total = self.impl.get_jobs(page, limit, search, sort)

        for info in infos:
            info['next_run_time'] = utc_timestamp_to_datetime(info['next_run_time']).astimezone(self.tz)
            self.datetime_to_str(info)

        return GetSuccess(
            msg='get cron job list success!',
            data={
                'items': infos,
                'total': total,
                'page': page,
                'limit': limit
            })



import re

from flask import g
from wtforms import StringField

from bspider.core.api import BaseForm, ParamRequired, Forbidden
from bspider.utils.exceptions import ModuleError
from bspider.utils.tools import find_class_name_by_content


def valid_code(content):
    description = 'default'
    pre_description = re.compile('@description=(.*?)\n').findall(content)
    if len(pre_description):
        description = pre_description[0]

    class_name, sub_class_name = find_class_name_by_content(content)

    # 非admin角色不能新增operation 类
    if class_name.lower().rfind('operation') != -1 and g.user.role != 'admin':
        raise Forbidden(msg='role has no permission!', errno=10004)

    code_type_map = {
        'BaseOperation': 'operation',
        'BasePipeline': 'pipeline',
        'BaseMiddleware': 'middleware',
        'BaseExtractor': 'extractor'
    }

    module_type = code_type_map.get(sub_class_name)
    if module_type is None:
        raise ModuleError('Unknown module type->%s(%s)' % (class_name, sub_class_name))

    # pre_exec
    try:
        exec(content)
    except Exception as e:
        raise ModuleError('module code has a exception:%s' % (e))

    return class_name, module_type, description, content


class AddForm(BaseForm):
    content = StringField(validators=[ParamRequired()])
    editor = StringField(validators=[ParamRequired()])

    def validate_content(self, value):
        value.data = valid_code(value.data)


class UpdateForm(BaseForm):
    content = StringField()
    editor = StringField()

    def validate_content(self, value):
        if value.data:
            value.data = valid_code(value.data)



from inspect import ismethod, isclass
from datetime import datetime

import six
from apscheduler.job import Job
from apscheduler.triggers.base import BaseTrigger
from apscheduler.util import convert_to_datetime, ref_to_obj, obj_to_ref, get_callable_name, check_callable_args
try:
    from collections.abc import Iterable, Mapping
except ImportError:
    from collections import Iterable, Mapping


class MySQLJob(Job):

    def __init__(self, scheduler, id=None, **kwargs):
        self._scheduler = scheduler
        self._jobstore_alias = None
        self._modify(id=id, **kwargs)

    def _modify(self, **changes):
        """
        Validates the changes to the Job and makes the modifications if and only if all of them
        validate.

        """
        approved = {}

        if 'id' in changes:
            value = changes.pop('id')
            if not isinstance(value, int):
                raise TypeError("id must be a int")
            approved['id'] = value

        if 'func' in changes or 'args' in changes or 'kwargs' in changes:
            func = changes.pop('func') if 'func' in changes else self.func
            args = changes.pop('args') if 'args' in changes else self.args
            kwargs = changes.pop(
                'kwargs') if 'kwargs' in changes else self.kwargs

            if isinstance(func, six.string_types):
                func_ref = func
                func = ref_to_obj(func)
            elif callable(func):
                try:
                    func_ref = obj_to_ref(func)
                except ValueError:
                    # If this happens, this Job won't be serializable
                    func_ref = None
            else:
                raise TypeError(
                    'func must be a callable or a textual reference to one')

            if not hasattr(self, 'name') and changes.get('name', None) is None:
                changes['name'] = get_callable_name(func)

            if isinstance(args, six.string_types) or not isinstance(args, Iterable):
                raise TypeError('args must be a non-string iterable')
            if isinstance(kwargs, six.string_types) or not isinstance(kwargs, Mapping):
                raise TypeError('kwargs must be a dict-like object')

            check_callable_args(func, args, kwargs)

            approved['func'] = func
            approved['func_ref'] = func_ref
            approved['args'] = args
            approved['kwargs'] = kwargs

        if 'name' in changes:
            value = changes.pop('name')
            if not value or not isinstance(value, six.string_types):
                raise TypeError("name must be a nonempty string")
            approved['name'] = value

        if 'misfire_grace_time' in changes:
            value = changes.pop('misfire_grace_time')
            if value is not None and (not isinstance(value, six.integer_types) or value <= 0):
                raise TypeError(
                    'misfire_grace_time must be either None or a positive integer')
            approved['misfire_grace_time'] = value

        if 'coalesce' in changes:
            value = bool(changes.pop('coalesce'))
            approved['coalesce'] = value

        if 'max_instances' in changes:
            value = changes.pop('max_instances')
            if not isinstance(value, six.integer_types) or value <= 0:
                raise TypeError('max_instances must be a positive integer')
            approved['max_instances'] = value

        if 'trigger' in changes:
            # 暂时先这样，后续支持多种定时任务的时候再进行修改
            trigger = changes.pop('trigger')
            if not isinstance(trigger, BaseTrigger):
                raise TypeError('Expected a trigger instance, got %s instead' %
                                trigger.__class__.__name__)

            # 每次更新trigger就重新计算下次执行时间
            now = datetime.now(self._scheduler.timezone)
            changes['next_run_time'] = trigger.get_next_fire_time(None, now)

            approved['trigger'] = trigger

        if 'executor' in changes:
            value = changes.pop('executor')
            if not isinstance(value, six.string_types):
                raise TypeError('executor must be a string')
            approved['executor'] = value

        if 'next_run_time' in changes:
            value = changes.pop('next_run_time')
            approved['next_run_time'] = convert_to_datetime(value, self._scheduler.timezone,
                                                            'next_run_time')
        # 新增两个属性 status, description
        if 'status' in changes:
            value = changes.pop('status')
            if value not in [0, 1]:
                raise TypeError("name status be 0 (closed) or 1(open)")
            approved['status'] = value

        if 'description' in changes:
            value = changes.pop('description')
            if not value or not isinstance(value, six.string_types):
                raise TypeError("description must be a nonempty string")
            approved['description'] = value

        if 'cron_type' in changes:
            value = changes.pop('cron_type')
            if not value or value not in ('operation', 'crawl'):
                raise TypeError("cron type must be operation or crawl")
            approved['type'] = value

        if changes:
            raise AttributeError('The following are not modifiable attributes of Job: %s' %
                                 ', '.join(changes))

        for key, value in six.iteritems(approved):
            setattr(self, key, value)

    def __getstate__(self):
        # Don't allow this Job to be serialized if the function reference could not be determined
        if not self.func_ref:
            raise ValueError(
                'This Job cannot be serialized since the reference to its callable (%r) could not '
                'be determined. Consider giving a textual reference (module:function name) '
                'instead.' % (self.func,))

        # Instance methods cannot survive serialization as-is, so store the "self" argument
        # explicitly
        if ismethod(self.func) and not isclass(self.func.__self__):
            args = (self.func.__self__,) + tuple(self.args)
        else:
            args = self.args

        return {
            'version': 1,
            'id': self.id,
            'func': self.func_ref,
            'trigger': self.trigger,
            'executor': self.executor,
            'args': args,
            'kwargs': self.kwargs,
            'name': self.name,
            'misfire_grace_time': self.misfire_grace_time,
            'coalesce': self.coalesce,
            'max_instances': self.max_instances,
            'next_run_time': self.next_run_time,
            # 新增两个属性 status, description
            'status': self.status,
            'description': self.description,
            # 新增属性type
            'type': self.cron_type
        }

    def __setstate__(self, state):
        if state.get('version', 1) > 1:
            raise ValueError('Job has version %s, but only version 1 can be handled' %
                             state['version'])

        self.id = state['id']
        self.func_ref = state['func']
        self.func = ref_to_obj(self.func_ref)
        self.trigger = state['trigger']
        self.executor = state['executor']
        self.args = state['args']
        self.kwargs = state['kwargs']
        self.name = state['name']
        self.misfire_grace_time = state['misfire_grace_time']
        self.coalesce = state['coalesce']
        self.max_instances = state['max_instances']
        self.next_run_time = state['next_run_time']
        # 新增两个属性 status, description
        self.status = state['status']
        self.description = state['description']
        # 新增属性type
        self.cron_type = state['type']

    def __repr__(self):
        return '<MySQLJob (id=%s name=%s)>' % (self.id, self.name)



import asyncio
import sys
import traceback

from bspider.config import FrameSettings
from bspider.scheduler.async_scheduler import AsyncScheduler
from bspider.utils.database import MysqlClient
from bspider.utils.sign import Sign


class SchedulerMonitor(object):

    def __init__(self, log, log_fn):
        self.frame_settings = FrameSettings()
        self.project_table = self.frame_settings['PROJECT_TABLE']
        self.__mysql_client = MysqlClient.from_settings(self.frame_settings['WEB_STUDIO_DB'])
        self.projects = {}
        self.log = log
        self.log_fn = log_fn

    def get_projects(self):
        sql = f'select `id`, `name`, `rate` from {self.project_table} where `status`=1'
        return self.__mysql_client.select(sql)

    async def sync_config(self):
        while True:
            self.log.info('start sync project!')
            tmp_projects = dict()
            try:
                infos = self.get_projects()
                self.log.debug(f'read info success: {infos}')
                for info in infos:
                    project_obj = self.projects.get(info['id'])
                    cur_sign = Sign(project_name=info['name'], project_id=info['id'], rate=info['rate'])
                    if project_obj is None or project_obj.sign != cur_sign:
                        tmp_projects[info['id']] = self.get_work_obj(
                            info['id'], info['name'], rate=info['rate'], sign=cur_sign)
                    else:
                        tmp_projects[info['id']] = project_obj
                self.projects = tmp_projects
                self.log.info('sync project success with {} projects'.format(len(self.projects)))
            except Exception as e:
                tp, msg, tb = sys.exc_info()
                e_msg = ''.join(traceback.format_exception(tp, msg, tb))
                self.log.error(f'server exec:{e_msg}')
                self.log.error(f'sync project rate error:{e}')
            await asyncio.sleep(4)

    def get_work_obj(self, project_id, project_name: str, rate: int, sign: Sign):
        return AsyncScheduler(project_id, project_name, rate, sign, self.log_fn)

    def close(self):
        pass



"""解析器监听器"""
from bspider.config.default_settings import EXCHANGE_NAME
from bspider.core import BaseMonitor, Project
from bspider.parser.async_parser import AsyncParser
from bspider.utils.sign import Sign


class ParserMonitor(BaseMonitor):
    exchange = EXCHANGE_NAME[2]

    def get_work_obj(self, project: Project, sign: Sign) -> AsyncParser:
        return AsyncParser(project, sign, self.log_fn)


import hashlib
import redis

from bspider.utils import singleton


@singleton
class RedisPoolFactory(object):
    """Redis Connection Pool"""

    def __init__(self):
        self.__pools = dict()

    @staticmethod
    def __make_hash(config):
        hashstr = ','.join([f'{k}:{v}' for k, v in config.items()])
        hashcode = hashlib.md5(hashstr.encode('utf8')).hexdigest()
        return hashcode

    def get_pool(self, redis_config: dict, max_connections: int) -> redis.Redis:
        """
        针对不同的redis配置缓存不同的连接池
        :param redis_config:
        host: ip or url
        port: int
        db: int
        password: None or str
        :max_connections: int default 5
        :return: conn pool
        """
        hashcode = self.__make_hash(redis_config)
        if hashcode in self.__pools:
            return redis.Redis(connection_pool=self.__pools[hashcode])
        else:
            self.__pools[hashcode] = redis.ConnectionPool(
                host=redis_config['host'],
                port=redis_config['port'],
                db=redis_config.get('db', 0),
                password=redis_config.get('password', None),
                decode_responses=True,
                max_connections=max_connections
            )
            return redis.Redis(connection_pool=self.__pools[hashcode])

# 这里假装这是个类 -_-
def RedisClient(max_connections: int = 5, **redis_config) -> redis.Redis:
    """
    返回redis的连接句柄
    :param redis_config:
        host: ip or url
        port: int
        db: int
        password: None or str
        max_connections: int default 5
    :return: redis handler
    """
    return RedisPoolFactory().get_pool(redis_config, max_connections)



"""
这里实例化一个对象，启用单独的线程异步获取CPU\DESK\MEM 使用率、系统配置
"""

import psutil

from bspider.config import FrameSettings


class System(object):
    """
    #### 后续增加异步获取CPU等使用率
    """
    master_ip = FrameSettings()['MASTER']['ip']
    agent_ip = FrameSettings().get('AGENT', {}).get('ip')
    ip_msg = f'{master_ip}->{agent_ip}' if agent_ip else master_ip

    @property
    def cpu_percent(self):
        return psutil.cpu_percent()

    @property
    def mem_percent(self):
        return psutil.virtual_memory().percent

    @property
    def disk_percent(self):
        return psutil.disk_usage('/').percent

    @property
    def cpu_num(self):
        return psutil.cpu_count(logical=True)

    @property
    def mem_size(self):
        """返回字节大小"""
        return psutil.virtual_memory().total/1024/1024/1024

    @property
    def disk_size(self):
        return psutil.disk_usage('/').total/1024/1024/1024


class WorkerProcess():
    pass


if __name__ == '__main__':
    print(System().cpu_percent)

    ps_info = {
        'CPU COUNT': psutil.cpu_count(),
        'CPU PERCENT': psutil.cpu_percent(),
        'MEM SIZE': psutil.virtual_memory().total/1024/1024/1024,
        'MEM PERCENT': psutil.virtual_memory().percent,
        'DISK SIZE': psutil.disk_usage('/').total/1024/1024/1024,
        'DISK PERCENT': psutil.disk_usage('/').percent
    }

    for key, value in ps_info.items():
        print(f'{key}:{value}')





from flask import Blueprint

from bspider.core.api import auth
from bspider.core.api import NotFound
from .validators import PageForm
from .validators.user_forms import LoginForm, RegisterForm, UpdateForm
from bspider.master.service.user import UserService

user = Blueprint('user_bp', __name__)

user_service = UserService()


@user.route('/login', methods=['POST'])
def login():
    form = LoginForm()
    client = form.client_type.data
    if client == 'identity':
        return user_service.login_by_identity(form.identity.data, form.password.data)
    else:
        return NotFound(msg='unknow client type')


@user.route('/user', methods=['POST'])
@auth.login_required
def add_user():
    form = RegisterForm()
    return user_service.add_user(**form.to_dict())


@user.route('/user/<int:user_id>', methods=['DELETE'])
@auth.login_required
def delete_user(user_id):
    return user_service.remove_user(user_id)


@user.route('/user/<int:user_id>', methods=['PATCH'])
@auth.login_required
def update_user(user_id):
    form = UpdateForm()
    return user_service.update_user(user_id, **form.to_dict())


@user.route('/user', methods=['GET'])
@auth.login_required
def get_users():
    form = PageForm()
    return user_service.get_users(int(form.page.data), int(form.limit.data), form.search.data, form.sort.data)


@user.route('/user/<int:user_id>', methods=['GET'])
@auth.login_required
def get_user(user_id):
    return user_service.get_user(user_id)



from functools import wraps

import pika

from bspider.config import FrameSettings
from bspider.utils import singleton


def retry(func):
    """重试失败的操作 最大三次"""
    @wraps(func)
    def _retry(self, *args, **kwargs):
        for i in range(3):
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                self.close()
                self.__init__(self.mq_config)

    return _retry


@singleton
class RabbitMQClient(object):

    def __init__(self, mq_config):
        self.mq_config = mq_config
        if 'username' in mq_config and 'password' in mq_config:
            pika.ConnectionParameters()
            credentials = pika.PlainCredentials(mq_config['username'], mq_config['password'])
        else:
            credentials = None
        self.connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=mq_config['host'],
                port=mq_config['port'],
                credentials=credentials,
                virtual_host=mq_config.get('virtual_host', None),
            )
        )
        self.channel = self.connection.channel()
        self.queue_arg = FrameSettings()['QUEUE_ARG']

    def close(self):
        try:
            self.channel.close()
            self.connection.close()
        except Exception as e:
            return e

    # Exchange API

    @retry
    def exchange_declare(self, exchange: str):
        self.channel.exchange_declare(exchange=exchange, exchange_type='topic', durable=True)

    # Queue API
    @retry
    def get_queue_message_count(self, queue):
        """
        获取队列的msg数量
        :param queue:
        :return:
        """
        result = self.declare_queue(queue)
        return result.method.message_count

    @retry
    def queue_declare(self, queue):
        """声明一个队列 新版pika"""
        return self.channel.queue_declare(
            queue=queue, durable=True, arguments=self.queue_arg)

    @retry
    def queue_delete(self, queue):
        """删除一个队列-- 这个方法会直接删除队列，很危险"""
        return self.channel.queue_delete(queue)

    @retry
    def queue_bind(self, queue, exchange, routing_key):
        """绑定一个队列"""
        self.channel.queue_bind(queue=queue, exchange=exchange, routing_key=routing_key)

    @retry
    def send_msg(self, exchange, routing_key, body, priority=3):
        """默认使用消息持久化"""
        properties = pika.BasicProperties(delivery_mode=2, priority=priority)
        self.channel.basic_publish(exchange=exchange, routing_key=routing_key, body=body, properties=properties)
        return True

    @retry
    def recv_msg(self, queue):
        """拉一条新的msg"""
        method, properties, body = self.channel.basic_get(queue=queue)
        # 队列是否为空
        if method is not None and isinstance(body, bytes):
            return method.delivery_tag, body.decode()
        return None, None

    @retry
    def report_acknowledgment(self, tag):
        """回ack给mq"""
        if isinstance(tag, int):
            self.channel.basic_ack(delivery_tag=tag)

    @retry
    def report_unacknowledgment(self, tag):
        if isinstance(tag, int):
            self.channel.basic_nack(delivery_tag=tag)



"""
上报下载结果 -> log
监听配置变化 -> sqlite
"""
import asyncio
import base64
import json

import signal
import zlib

from bspider.http import Response
from bspider.utils.database import AioMysqlClient
from bspider.utils.logger import LoggerPool

from .broker import RabbitMQBroker
from .api import BaseImpl


class BaseManager(object):
    manager_type = ''

    def __init__(self, unique_tag, monitor_cls, coro_num):
        # 记录协程数
        self.coro_num = coro_num

        if self.manager_type == 'scheduler':
            self.log = LoggerPool().get_logger(
                key=unique_tag,
                fn=f'{self.manager_type}',
                module=self.manager_type,
                name=unique_tag)
            self.broker = RabbitMQBroker(self.log, 10)
            self.monitor = monitor_cls(self.log, self.manager_type)
        else:
            self.log = LoggerPool().get_logger(
                key=unique_tag,
                fn=f'{self.manager_type}-{unique_tag}',
                module=self.manager_type,
                name=unique_tag)
            self.broker = RabbitMQBroker(self.log, self.coro_num)
            self.monitor = monitor_cls(self.log, f'{self.manager_type}-{unique_tag}', self.broker.mq_client)

        self.log.info('manager init success')
        # 注册 任务中间人

        if self.manager_type == 'downloader':
            self.status_table = self.broker.frame_settings['DOWNLOADER_STATUS_TABLE']
        elif self.manager_type == 'parser':
            self.status_table = self.broker.frame_settings['PARSER_STATUS_TABLE']

        if self.manager_type != 'scheduler':
            self.mysql_client = AioMysqlClient(self.broker.frame_settings['WEB_STUDIO_DB'])

        # 注册信号量保证程序安全退出
        self.sign = signal.signal(signal.SIGTERM, handler=self.close)
        self.is_close = False

    def run(self):
        tasks = [asyncio.ensure_future(self.monitor.sync_config())]
        for i in range(self.coro_num):
            tasks.append(asyncio.ensure_future(self.do_work()))
        self.loop = asyncio.get_event_loop()
        try:
            self.loop.run_until_complete(asyncio.wait(tasks))
        except (KeyboardInterrupt, SystemExit):
            self.loop.run_until_complete(asyncio.sleep(1))
        finally:
            self.close('SIGINT', '')

    def close(self, signum, frame):
        """进程结束回调"""
        if not self.is_close:
            self.log.info('recv signal:{} {}'.format(signum, frame))
            try:
                self.loop.stop()
                self.loop.close()
            except Exception as e:
                # 吞掉异常
                self.log.debug(f'close event loop! {e}')
            # self.monitor.close()
            self.log.info('have fun Bye!')
            exit()
        self.is_close = True

    async def do_work(self):
        raise NotImplementedError

    async def _save_success_result(self,
                                   response: Response,
                                   project_name: str,
                                   project_id: int):
        data = {
            'project_id': project_id,
            'project_name': project_name,
            'sign': response.sign,
            'method': response.method,
            'url': response.url,
            'status': response.status,
            'url_sign': response.url,
        }
        if hasattr(response.request, 'data'):
            data['data'] = json.dumps(response.request.data)
        else:
            data['data'] = None
        fields, values = BaseImpl.make_fv(data)
        sql = f'insert into {self.status_table} set {fields};'.replace('`url_sign`=%s', '`url_sign`=md5(%s)')
        await self.mysql_client.insert(sql, values)
        self.log.info('send success info [{}]'.format(fields % values))

    async def _save_error_result(self,
                                 response: Response,
                                 project_name: str,
                                 project_id: int,
                                 exception: str):
        request = response.request
        data = {
            'project_id': project_id,
            'project_name': project_name,
            'sign': request.sign,
            'method': request.method,
            'url': request.url,
            'status': response.status,
            'url_sign': request.url,
            'exception': exception,
            'response': base64.b64encode(zlib.compress(json.dumps(response.dumps()).encode()))
        }
        if hasattr(request, 'data'):
            data['data'] = json.dumps(request.data)
        else:
            data['data'] = None
        fields, values = BaseImpl.make_fv(data)
        sql = f'insert into {self.status_table} set {fields};'
        await self.mysql_client.insert(sql.replace('`url_sign`=%s', '`url_sign`=md5(%s)'), values)
        self.log.info('send error info [{}]'.format(fields % values))



import json

from flask import request
from werkzeug.exceptions import HTTPException


class APIException(HTTPException):
    code = 500  # 通用错误响应
    msg = 'sorry, server make a mistake'
    errno = -1
    data = None

    def __init__(self, code=None, msg=None, errno=None, data=None):
        if code:
            self.code = code
        if msg:
            self.msg = msg
        if errno:
            self.errno = errno
        self.data = data
        super().__init__(description=self.msg)

    def get_body(self, environ=None):
        req_uri = '{} {}'.format(request.method, str(request.full_path).split('?')[0])
        body = {
            'msg': self.msg,
            'errno': self.errno,
            'request': req_uri,
        }
        if self.data is not None:
            body['data'] = self.data
        return json.dumps(body)

    def get_headers(self, environ=None):
        return [("Content-Type", "application/json")]


class ClientTypeError(APIException):
    code = 400  # 参数错误
    msg = 'client is invalid'
    errno = 10001


class NotFound(APIException):
    code = 404
    msg = 'not found'
    errno = 101


class ParameterException(APIException):
    code = 400
    msg = 'invalid parameter'
    errno = 100


class AuthFailed(APIException):
    code = 401
    msg = 'Unauthorized'
    errno = 10001


class Forbidden(APIException):
    code = 403
    msg = 'forbidden, op failed'
    errno = 10004


class PreconditionFailed(APIException):
    code = 412
    msg = 'Precondition Failed'
    errno = 102


class Conflict(APIException):
    """操作异常"""
    code = 409
    msg = 'conflict'
    errno = 103



from bspider.core.custom_module import BaseCustomModule
from bspider.http import Request, Response


class BaseMiddleware(BaseCustomModule):

    def process_request(self, request: Request):
        """
        执行下载前的操作
        返回Response类型，作为response完成下载，跳过后面的下载阶段，直接进入下载后的中间件
        返回空或其他返回值 继续下一个中间件
        此为协程方式也可通过常规方式调用
        def process_request(self, request)
        """
        return None


    def process_response(self, request: Request, response: Response):
        """
        执行下载后的操作，各个中间件进行重写
        返回Response类型，正常，继续下一个中间件
        返回None类型或其他返回值，丢弃，后面不再处理
        """
        return response


    def process_exception(self, request: Request, e, response: Response):
        """
        这里捕获下载时发生的异常，不包括中间件异常
        返回Response类型，正常，继续下一个中间件
        返回None类型或其他返回值，丢弃，后面不再处理
        """
        return response



from bspider.agent import log
from bspider.core.api import BaseService, GetSuccess, PostSuccess, DeleteSuccess, PatchSuccess
from bspider.core import AgentCache


class ProjectService(BaseService):

    def __init__(self):
        self.cache = AgentCache()

    def add_project(self, project_id, name, config, rate, status):
        self.cache.set_project(project_id, name, config, rate, status)
        log.info(f'add project:project_id->{project_id} project_name->{name} {status} success')
        return PostSuccess()

    def get_projects(self):
        return GetSuccess(data=self.cache.get_projects())

    def get_project(self, project_id):
        return GetSuccess(data=self.cache.get_project(project_id))

    def update_project(self, project_id, changes):
        self.cache.update_project(project_id, changes)
        log.info(f'update project:{project_id} success')
        return PatchSuccess()

    def delete_project(self, project_id):
        self.cache.delete_project(project_id)
        log.info(f'delete project:{project_id} success')
        return DeleteSuccess()



import logging
import os
import sys
import re
from logging.handlers import TimedRotatingFileHandler

from bspider.utils import singleton
from bspider.config import FrameSettings
from bspider.utils.conf import PLATFORM_PATH_ENV
from bspider.utils.sign import Sign

from .formatter import get_stream_formatter, get_file_formatter


@singleton
class LoggerPool(object):
    """日志句柄池"""

    def __init__(self):
        self.__pool = {}
        self.frame_settings = FrameSettings()
        if self.frame_settings['GLOBAL_PATTERN'] == 'produce':
            self.__handle_func = self.__set_file_handler
        else:
            self.__handle_func = self.__set_stream_handler

    def __set_stream_handler(self, log, level='DEBUG', **kwargs):
        log_handler = logging.StreamHandler(stream=sys.stderr)
        log_handler.suffix = "%Y%m%d"
        log_handler.extMatch = re.compile(r"^\d{4}\d{2}\d{2}$")
        log_formatter = get_stream_formatter(**kwargs)
        log_handler.setFormatter(log_formatter)
        log.addHandler(log_handler)
        if level == 'INFO':
            log.setLevel(logging.INFO)
        elif level == 'WARNING':
            log.setLevel(logging.WARNING)
        elif level == 'DEBUG':
            log.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)
        return log

    def __set_file_handler(self, log, fn, level='INFO', **kwargs):
        log_path = os.path.join(os.environ[PLATFORM_PATH_ENV], 'log', f'{fn}.log')
        log_handler = TimedRotatingFileHandler(log_path, when="midnight", backupCount=10)
        log_handler.suffix = "%Y%m%d"
        log_formatter = get_file_formatter(**kwargs)
        log_handler.setFormatter(log_formatter)
        log.addHandler(log_handler)
        if level == 'INFO':
            log.setLevel(logging.INFO)
        elif level == 'WARNING':
            log.setLevel(logging.WARNING)
        elif level == 'DEBUG':
            log.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)
        return log

    def get_logger(self, key, fn, **kwargs) -> logging.Logger:
        """如果是downloader、parser模块控制终端输出到指定log文件"""
        cur_sign = Sign(**kwargs)
        if key not in self.__pool:
            log_handler = self.__handle_func(
                log=logging.getLogger(key),
                fn=fn,
                level=self.frame_settings['LOGGER_LEVEL'],
                **kwargs)
            self.__pool[key] = (cur_sign, log_handler)
            return log_handler

        pre_sign, log_handler = self.__pool[key]
        if pre_sign == cur_sign:
            return log_handler
        else:
            for handler in log_handler.handlers:
                handler.close()
                log_handler.removeHandler(handler)

            log_handler = self.__handle_func(
                log=logging.getLogger(key),
                fn=fn,
                level=self.frame_settings['LOGGER_LEVEL'],
                **kwargs)

            self.__pool[key] = (cur_sign, log_handler)
            return log_handler

        # if kwargs.get('module') in ('parser', 'downloader'):
        #     log_handler = self.__set_file_handler(logging.getLogger(unique_key), self.frame_settings['LOGGER_LEVEL'], key=key,
        #                                           **kwargs)
        # else:



from flask import Blueprint

from bspider.core.api import auth
from .validators import PageForm
from .validators.project_form import UpdateForm, AddForm
from bspider.master.service.project import ProjectService

project = Blueprint('project_bp', __name__)

project_service = ProjectService()


@project.route('/project/<int:project_id>', methods=['GET'])
@auth.login_required
def get_project(project_id):
    return project_service.get(project_id)


@project.route('/project', methods=['GET'])
@auth.login_required
def get_projects():
    form = PageForm()
    return project_service.gets(**form.to_dict())


@project.route('/project/<int:project_id>', methods=['DELETE'])
@auth.login_required
def delete_project(project_id):
    return project_service.delete(project_id)


@project.route('/project/<int:project_id>', methods=['PATCH'])
@auth.login_required
def update(project_id):
    form = UpdateForm()
    return project_service.update(project_id, form.to_dict())


@project.route('/project', methods=['POST'])
@auth.login_required
def add():
    form = AddForm()
    return project_service.add(**form.to_dict())



from .sqlite3_client import SqlLite3Client
from .redis_client import RedisClient
from .mysql.client import MysqlClient
from .mysql.async_client import AioMysqlClient
from .mysql import prepare_insert_sql




import os
from os import remove
from shutil import rmtree

from bspider.commands import BSpiderCommand
from bspider.utils.conf import PLATFORM_NAME_ENV, BSPIDER_CONFIG_PATH_ENV, PLATFORM_PATH_ENV
from bspider.utils.exceptions import UsageError


class Command(BSpiderCommand):

    def syntax(self):
        return "<platform_name>"

    def short_desc(self):
        return "remove current platform"

    def run(self, args, opts):
        if len(args) not in (1, 2):
            raise UsageError('args error')

        platform_name = args[0]
        platform_path = os.environ[PLATFORM_PATH_ENV]


        if not os.environ.get(PLATFORM_NAME_ENV):
            print(f'Error: platform:{platform_name} is not exists')
            self.exitcode = 1
            return

        # 停止supervisor
        try:
            with open(os.path.join(platform_path, '.cache', 'supervisord.pid')) as f:
                pid = f.read().strip()
            print(os.popen(f'kill {pid}').read().strip())
            print('Stop supervisor process success')
        except Exception:
            print(f'warning: supervisor process is not running')


        print('Clean platform file ...')
        # 删除文件
        rmtree(os.environ[PLATFORM_PATH_ENV], ignore_errors=True)
        remove(os.environ[BSPIDER_CONFIG_PATH_ENV])
        print(f"BSpider platform '{platform_name}' remove success")



import datetime
from apscheduler.jobstores.base import BaseJobStore, JobLookupError
from apscheduler.triggers.cron import CronTrigger

from apscheduler.util import utc_timestamp_to_datetime, datetime_to_utc_timestamp

from bspider.bcron.job import MySQLJob
from bspider.utils.database import MysqlClient


class MySQLJobStore(BaseJobStore):
    """
    Stores jobs in a database table using pymysql.
    if table doesn't exist in the database, you must build it by yourself.
    :param mysql_client: a pymysql conn obj
    :param str tablename: name of the table to store jobs in
    """

    TABLE_FIELDS = ('`id`', '`project_id`', '`code_id`', '`trigger`', '`trigger_type`', '`type`',
                    '`next_run_time`', '`executor`', '`func`', '`status`', '`description`')

    def __init__(self, mysql_client: MysqlClient, tz, log, table_name='bspider_cronjob'):
        super().__init__()
        # 重构获取一个mysql连接池
        self.mysql_client = mysql_client
        self.table_name = table_name
        self.tz = tz
        self.log = log

    def _reconstitute_job(self, job_state: dict) -> MySQLJob:
        """
        mysql返回的字典解析成 job对象
        :param job_state:
        :return:
        """
        job_state['jobstore'] = self
        job_state['name'] = '{project_id}-{code_id}'.format(**job_state)
        if job_state['trigger_type'] == 'cron':
            job_state['trigger'] = CronTrigger.from_crontab(job_state['trigger'])
        job_state['misfire_grace_time'] = None
        job_state['coalesce'] = True
        job_state['max_instances'] = 1
        job_state['args'] = []
        job_state['kwargs'] = {
            'project_id': job_state['project_id'],
            'code_id': job_state['code_id'],
            'type': job_state['type']
        }
        job_state['id'] = job_state['id']
        job_state['next_run_time'] = datetime.datetime.fromtimestamp(job_state['next_run_time'], self.tz)
        job = MySQLJob.__new__(MySQLJob)
        job.__setstate__(job_state)
        job._scheduler = self._scheduler
        job._jobstore_alias = self._alias
        return job

    def __make_fv(self, job: MySQLJob) -> tuple:
        fields = ','.join([' %s=%%s ' % (key) for key in self.TABLE_FIELDS])
        project_id, code_id = job.name.split('-')

        trigger_type = job.trigger.__str__().split('[')[0]

        if trigger_type == 'cron':
            options = {f.name: f.__str__() for f in job.trigger.fields if not f.is_default}
            trigger = '{minute} {hour} {day} {month} {day_of_week}'.format(**options)

        else:
            trigger = ''

        values = (job.id, project_id, code_id, trigger, trigger_type, job.cron_type,
                  job.next_run_time if isinstance(job.next_run_time, float) else datetime_to_utc_timestamp(job.next_run_time),
                  job.executor, job.func_ref, job.status, job.description)

        return (fields, values)

    def add_job(self, job: MySQLJob):
        """因为拆分问题，增加job 的操作交给API模块完成"""
        update = f"update {self.table_name} set `status`=%s where `id` = '{job.id}';"
        self.mysql_client.update(update, (0,))
        self.log.info(f'sync job:{job.name} success')

    def lookup_job(self, job_id):
        sql = f"select * from {self.table_name} where `id` = '{job_id}'"
        job_state = self.mysql_client.select(sql)
        return self._reconstitute_job(job_state[0]) if len(job_state) else None

    def get_due_jobs(self, now):
        timestamp = datetime_to_utc_timestamp(now)
        return self._get_jobs(timestamp)

    def get_next_run_time(self):
        selectable = f'select `next_run_time` from {self.table_name} ' \
            f'where `next_run_time` is not null ' \
            f'order by `next_run_time` limit 1'
        info = self.mysql_client.select(selectable)

        return utc_timestamp_to_datetime(info[0]['next_run_time']) if len(info) else None

    def get_all_jobs(self):
        jobs = self._get_jobs()
        self._fix_paused_jobs_sorting(jobs)
        return jobs

    def update_job(self, job: MySQLJob):
        fields, values = self.__make_fv(job)
        update = f"update {self.table_name} set {fields} where `id` = '{job.id}';"
        self.log.debug(update % values)
        result = self.mysql_client.update(update, values)
        if result == 0:
            self.log.debug(f'job: {job.name} is nothing to change！')
        else:
            self.log.debug(f'job: {job.name} is update success！')

    def remove_job(self, job_id):
        delete = f'DELETE FROM {self.table_name} WHERE `id` = {job_id};'
        result = self.mysql_client.delete(delete)
        if result == 0:
            raise JobLookupError(job_id)

    def remove_all_jobs(self):
        delete = f'DELETE FROM {self.table_name};'
        self.mysql_client.delete(delete)

    def shutdown(self):
        self.log.warning('mysql job store shutdown the conn pool!')

    def _get_jobs(self, timestamp=None):
        jobs = []
        failed_job_ids = set()
        fields = ','.join(self.TABLE_FIELDS)

        if timestamp:
            selectable = f'select {fields} from {self.table_name} ' \
                f'where `next_run_time` <= {timestamp}' \
                f'order by `next_run_time`;'
        else:
            selectable = f'select {fields} from {self.table_name} ' \
                f'order by `next_run_time`;'

        for row in self.mysql_client.select(selectable):
            try:
                jobs.append(self._reconstitute_job(row))
            except BaseException:
                self._logger.exception('unable to restore job "%s" -- removing it', row['id'])
                failed_job_ids.add(row['id'])
        return jobs

    def __repr__(self):
        return '<%s (engine=%s)>' % (self.__class__.__name__, 'MySQL')



from flask import request
from wtforms import Form, IntegerField
from wtforms.validators import DataRequired

from .exception import ParameterException


class BaseForm(Form):

    def __init__(self):
        tmp = request.args.to_dict()
        if request.json:
            for key, values in request.json.items():
                tmp[key] = values
        if request.form:
            for key, values in request.form.items():
                tmp[key] = values
        super().__init__(data=tmp)
        self.__validate_for_api()

    def __validate_for_api(self):
        if not super().validate():
            raise ParameterException(msg=self.errors)

    def to_dict(self):
        result = {}
        for name, field in self._fields.items():
            if field is not None and field.data is not None:
                if isinstance(field, IntegerField):
                    field.data = int(field.data)
                result[name] = field.data

        if len(result):
            return result
        raise ParameterException(msg='all arguments is Invalid')


class ParamRequired(DataRequired):
    """修复wtf DataRequired int=0 时识别失败的bug"""
    def __call__(self, form, field):
        if isinstance(field.data, int) and isinstance(field, IntegerField):
            return
        super().__call__(form, field)




from flask import Blueprint, g

from bspider.core.api import auth
from bspider.master.service.tools import ToolsService

from .validators.tools_forms import GetCodeListForm, ValidateForm, GetCrawlDetailForm

tools = Blueprint('tools_bp', __name__)

tools_service = ToolsService()


@tools.route('/node-list', methods=['GET'])
@auth.login_required
def node_list():
    """返回节点列表"""
    return tools_service.get_node_list()


@tools.route('/code-list', methods=['GET'])
@auth.login_required
def code_list():
    """返回节点列表"""
    form = GetCodeListForm()
    return tools_service.get_code_list(form.type.data)


@tools.route('/validate/<string:valid_type>', methods=['GET'])
@auth.login_required
def validate(valid_type):
    """返回节点列表"""
    form = ValidateForm()
    return tools_service.validate(valid_type, form.data.data)


@tools.route('/parser/exception/<int:project_id>', methods=['GET'])
@auth.login_required
def parser_exception(project_id):
    """返回节点列表"""
    return tools_service.get_parser_exception(project_id)


@tools.route('/downloader/exception/<int:project_id>', methods=['GET'])
@auth.login_required
def downloader_exception(project_id):
    """返回节点列表"""
    return tools_service.get_downloader_exception(project_id)


@tools.route('/crawl-detail', methods=['GET'])
@auth.login_required
def crawl_detail():
    """返回节点列表"""
    form = GetCrawlDetailForm()
    return tools_service.get_crawl_detail(**form.to_dict())


@tools.route('/node-detail', methods=['GET'])
@auth.login_required
def node_detail():
    """返回节点列表"""
    return tools_service.get_node_detail()


@tools.route('/exception-project', methods=['GET'])
@auth.login_required
def exception_project():
    """返回节点列表"""
    return tools_service.get_exception_project()

@tools.route('/current_user', methods=['GET'])
@auth.login_required
def get_current_user():
    return tools_service.get_current_user(g.user.user_id)



from setuptools import setup, find_packages

extras_require = {}

setup(
    name='bspider',
    version=open('./bspider/VERSION').read().strip(),
    description='A high-level distributed crawling framework Git: https://github.com/littlebai3618/bspider',
    author='baii',
    author_email='2624386844@qq.com',
    license='BSD',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['bspider = bspider.cmdline:execute']
    },
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    install_requires=[
        'aiohttp==3.6.2',
        'aiomysql==0.0.20',
        'aiormq==2.7.2',
        'APScheduler==3.6.1',
        'asn1crypto==0.24.0',
        'async-timeout==3.0.1',
        'attrs==19.1.0',
        'casbin==0.7.1',
        'certifi==2019.6.16',
        'cffi==1.12.3',
        'chardet==3.0.4',
        'Click==7.0',
        'cryptography==2.7',
        'DBUtils==1.3',
        'Flask==1.1.1',
        'Flask-Cors==3.0.8',
        'Flask-HTTPAuth==3.3.0',
        'Flask-WTF==0.14.2',
        'gevent==1.4.0',
        'greenlet==0.4.15',
        'gunicorn==19.9.0',
        'idna==2.8',
        'itsdangerous==1.1.0',
        'Jinja2==2.10.1',
        'lxml==4.4.1',
        'MarkupSafe==1.1.1',
        'meld3==1.0.2',
        'multidict==4.5.2',
        'pamqp==2.3.0',
        'pika==1.1.0',
        'psutil>=5.6.6',
        'pycparser==2.19',
        'PyMySQL==0.9.2',
        'pytz==2019.2',
        'requests==2.22.0',
        'simpleeval==0.9.8',
        'six==1.12.0',
        'supervisor==4.0.4',
        'tzlocal==2.0.0',
        'urllib3==1.25.3',
        'Werkzeug==0.15.5',
        'WTForms==2.2.1',
        'xpinyin==0.5.6',
        'yarl==1.3.0',
        'redis==3.3.11',
        'PyYAML==5.2',
        'voluptuous==0.11.7'
    ],
    extras_require=extras_require,
)



def prepare_insert_sql(table: str, data: dict, auto_update:bool = False, **kwargs) -> (str, tuple):
    """
    dict数据 -> sql
    :param table: mysql表名
    :param data: 插入的数据
    :param kwargs: 其他参数
        auto_update bool
        immutable_fields list：当auto_update = True时 生效，忽略一些参数的更新
    :return: sql, value
    """
    import io
    sbuf = io.StringIO()
    sbuf.write('INSERT INTO %s SET ' % table)
    sbuf.write(','.join([
        ' %s=%%s ' % (key) for key in data.keys() if data.get(key)
    ]))
    values = [data[key] for key in data.keys() if data.get(key)]

    if not auto_update:
        return sbuf.getvalue(), tuple(values)

    sbuf.write('ON DUPLICATE KEY UPDATE ')
    sbuf.write(','.join([
        ' %s=%%s ' % (key) for key in data.keys()
        if data.get(key) and key not in kwargs.get('immutable_fields', [])
    ]))

    values.extend([
        data[key] for key in data.keys()
        if data.get(key) and key not in kwargs.get('immutable_fields', [])
    ])
    return sbuf.getvalue(), tuple(values)



from bspider.utils.database import MysqlClient
from bspider.config import FrameSettings
from bspider.utils.tools import make_fields_values


class BaseImpl(object):
    frame_settings = FrameSettings()
    project_table = frame_settings['PROJECT_TABLE']
    code_table = frame_settings['CODE_STORE_TABLE']
    cron_table = frame_settings['CRON_JOB_STORE_TABLE']
    node_table = frame_settings['NODE_TABLE']
    user_table = frame_settings['USER_TABLE']
    p2c_table = frame_settings['P2C_TABLE']

    downloader_status_table = frame_settings['DOWNLOADER_STATUS_TABLE']
    parser_status_table = frame_settings['PARSER_STATUS_TABLE']

    worker_table = frame_settings['WORKER_TABLE']
    node_status_table = frame_settings['NODE_STATUS_TABLE']

    mysql_client = MysqlClient.from_settings(frame_settings['WEB_STUDIO_DB'])

    @staticmethod
    def make_fv(data: dict) -> tuple:
        """
        给定字典，返回fields 和 values
        :param info: dict
        :return:
        """
        return make_fields_values(data)

    @staticmethod
    def make_search(search: str) -> str:
        """
        给定字典，返回fields 和 values
        :param info: dict
        :return:
        """
        field = list()
        if len(search):
            for kv in search.split(','):
                k, v = kv.split('=')
                field.append('`{}` like \'%%{}%%\' '.format(k, v))
        return ' and '.join(field)

    def total_num(self, search, table_name):
        fields = self.make_search(search)
        if len(fields):
            sql = f"select count(1) as total from `{table_name}` where {fields}; "
        else:
            sql = f"select count(1) as total from `{table_name}`; "
        return self.mysql_client.select(sql)[0]['total']

# def remove(self, unique_value, unique_key='id'):
#     sql = f"update {self.table_name} set `status`=%s where `{unique_key}` = '{unique_value}';"
#     return self.handler.update(sql, (-1))
#
# def update(self, unique_value, data, unique_key='id'):
#     fields, values = self.make_fv(data)
#     sql = f"update {self.table_name} set {fields} where `{unique_key}` = '{unique_value}';"
#     return sql, values
#
# def add(self, data):
#     fields, values = self.make_fv(data)
#     sql = f"insert into {self.table_name} set {fields};"
#     return self.handler.insert(sql, values, lastrowid=True)



from bspider.config.default_settings import EXCHANGE_NAME
from bspider.core.api import BaseImpl
from bspider.utils.rabbitMQ import RabbitMQClient
from bspider.master import log


class ProjectImpl(BaseImpl):

    def __init__(self):
        self.__mq_client = RabbitMQClient(self.frame_settings['RABBITMQ_CONFIG'])

    def get_project(self, project_id):
        sql = f'select `id`, `name`, `status`, `type`, `group`, `description`, `editor`, `rate`, `config`, `create_time`, `update_time` ' \
              f'from {self.project_table} where `id`="{project_id}";'
        return self.mysql_client.select(sql)

    def get_projects(self, page, limit, search, sort):
        start = (page - 1) * limit
        fields = self.make_search(search)
        if len(fields):
            sql = f'select `id`, `name`, `status`, `type`, `group`, `description`, `editor`, `rate`, `create_time`, `update_time` ' \
                  f'from {self.project_table} where {fields} order by `id` {sort} limit {start},{limit};'
        else:
            sql = f'select `id`, `name`, `status`, `type`, `group`, `description`, `editor`, `rate`, `create_time`, `update_time` ' \
                  f'from {self.project_table} order by `id` {sort} limit {start},{limit};'
        log.debug(f'SQL:{sql}')
        return self.mysql_client.select(sql), self.total_num(search, self.code_table)

    def add_project(self, data):
        log.debug(data)
        fields, values = self.make_fv(data)
        sql = f"insert into {self.project_table} set {fields};"
        return sql, values

    def add_project_binds(self, cids, pid):
        values = ', '.join([f'({pid}, {cid})' for cid in cids])
        sql = f'replace into `{self.p2c_table}`(`project_id`, `customcode_id`) ' \
              f'values{values};'
        log.debug(sql)
        return sql,

    def update_project(self, unique_value, data, unique_key='id'):
        fields, values = BaseImpl.make_fv(data)
        sql = f"update {self.project_table} set {fields} where `{unique_key}` = '{unique_value}';"
        log.info(sql, values)
        return sql, values

    def get_module_id_by_name_and_type(self, code_name, code_type):
        sql = f'select `id` from {self.code_table} where `name`="{code_name}" and `type`="{code_type}"'
        return self.mysql_client.select(sql)

    def delete_project_binds(self, project_id):
        sql = f'delete from {self.p2c_table} where `project_id`={project_id};'
        return sql,

    def delete_cron_job(self, project_id: int):
        sql = f"update {self.cron_table} set `status`=%s where `project_id` = '{project_id}';"
        log.debug(f'SQL:{sql}')
        return sql, (3)

    def add_cron_job(self, data: dict) -> tuple:
        data['status'] = 1
        fields, values = self.make_fv(data)
        sql = f"insert into {self.cron_table} set {fields}"
        log.debug(f'SQL:{sql}')
        return sql, values, True

    def update_cron_job_by_project_id(self, project_id: int, data: dict) -> tuple:
        data['status'] = 2
        fields, values = self.make_fv(data)
        sql = f"update {self.cron_table} set {fields} where `project_id` = '{project_id}';"
        log.debug(f'SQL:{sql}')
        return sql, values

    def delete_project(self, project_id):
        sql = f'delete from {self.project_table} where `id`={project_id};'
        return sql,

    def bind_queue(self, project_id):
        for exchange in EXCHANGE_NAME:
            queue_name = '{}_{}'.format(exchange, project_id)
            self.__bind(exchange, queue_name, str(project_id))
        return True

    def __bind(self, exchange, queue_name, routing_key):
        self.__mq_client.exchange_declare(exchange)
        self.__mq_client.queue_declare(queue_name)
        self.__mq_client.queue_bind(queue_name, exchange, routing_key)

    def unbind_queue(self, project_id):
        for exchange in EXCHANGE_NAME:
            res = self.__mq_client.queue_delete(queue=f'{exchange}_{project_id}')
            log.debug(f'delete queue: {exchange}_{project_id} => res:{res}')
        return True

    def get_nodes(self):
        sql = f'select `ip` from {self.node_table} where `status` = 1'
        return [info['ip'] for info in self.mysql_client.select(sql)]



from bspider.utils.logger import LoggerPool

log = LoggerPool().get_logger(key='master', fn='master', module=f'master')
# 注册一个Auth模块的句柄
LoggerPool().get_logger(key='api_auth', fn='master', module='api_auth')




import json
import xmlrpc.client

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError
from flask import g

from bspider.config import FrameSettings
from bspider.core.api.auth.token import make_token
from bspider.utils.exceptions import RemoteOPError


class RemoteMixIn(object):
    # 可以优化为协程方式调用

    @staticmethod
    def request(url, method, params=None, data=None, token=None):
        if token:
            headers = dict(Authorization=f'Bearer {token}')
        else:
            headers = dict(Authorization=f'Bearer {g.user.token}')
        if data:
            headers['Content-Type'] = 'application/json'
            data = json.dumps(data)
        try:
            req = requests.request(method, url, headers=headers, data=data, params=params)
        except ConnectionError as e:
            raise RemoteOPError('Call Remote failed please check!, error:%s' % (e))
        return req


class AgentMixIn(RemoteMixIn):
    """
    master -> agent 主节点向子节点进行通信
    """
    base_url = 'http://{}:' + str(FrameSettings()['AGENT'].get('port', 5001)) + '{}'
    rpc_url = 'http://{username}:{password}@%s:{port}/RPC2'.format(**FrameSettings()['SUPERVISOR_RPC'])

    def op_stop_node(self, ip: str) -> bool:
        with xmlrpc.client.ServerProxy(self.rpc_url % ip) as rpc_server:
            if rpc_server.supervisor.stopProcess('agent:agent'):
                return True
        return False

    def op_start_node(self, ip: str) -> bool:
        with xmlrpc.client.ServerProxy(self.rpc_url % ip) as rpc_server:
            if rpc_server.supervisor.startProcess('agent:agent'):
                return True
        return False

    def op_get_node(self, ip: str) -> dict:
        result = {
            'description': f'Agent {ip} process is not exist!!!',
            'pid': 0,
            'status': -1
        }

        with xmlrpc.client.ServerProxy(self.rpc_url % ip) as rpc_server:
            process = rpc_server.supervisor.getAllProcessInfo()[0]
            if process['name'] == 'agent':
                result['description'] = process['description']
                result['pid'] = process['pid']
                if process['state'] == 20:
                    result['status'] = 1
                elif process['state'] == 0:
                    result['status'] = 0
                else:
                    result['description'] = 'Error see->{}'.format(process['stdout_logfile'])
                    result['pid'] = process['pid']
                return result
        return result

    def op_stop_worker(self, ip: str, worker_id: int, token: str = None) -> bool:
        url = self.base_url.format(ip, f'/worker/{worker_id}')
        req = self.request(url, method='DELETE', token=token)
        if req.status_code == 204:
            return True
        raise RemoteOPError('stop work error %s' % (req.json()['msg']))

    def op_start_worker(self, ip: str, worker_id: int, worker_type: str, coroutine_num: int, token: str = None) -> dict:
        url = self.base_url.format(ip, '/worker')
        data = {
            'worker_id': worker_id,
            'worker_type': worker_type,
            'coroutine_num': coroutine_num
        }
        req = self.request(url, method='POST', data=data, token=token)
        data = req.json()
        if data['errno'] == 0:
            return data['data']

    def op_get_worker(self, ip: str, worker_id: int, token: str = None) -> dict:
        url = self.base_url.format(ip, f'/worker/{worker_id}')
        req = self.request(url, method='GET', token=token)
        data = req.json()
        if data['errno'] == 0:
            return data['data']
        raise RemoteOPError('get worker error %s' % (req.json()['msg']))

    def __op_query(self, ip_list: list, method: str, uri: str, data: dict = None, token: str = None) -> (bool, dict):
        result = dict()
        if method == 'DELETE':
            for ip in ip_list:
                req = self.request(self.base_url.format(ip, uri), method=method, data=data, token=token)
                if not (req.status_code >= 200 and req.status_code <= 299):
                    result[ip] = req.json()['msg']
        else:
            for ip in ip_list:
                data = self.request(self.base_url.format(ip, uri), method=method, data=data, token=token).json()
                if data['errno'] != 0:
                    result[ip] = data['msg']
        return not len(result), result

    def op_add_project(self, ip_list: list, data: dict, token: str = None) -> (bool, dict):
        # project_id, name, config, rate, status
        return self.__op_query(ip_list, 'POST', '/project', data, token=token)

    def op_update_project(self, ip_list: list, project_id: int, data: dict, token: str = None) -> (bool, dict):
        return self.__op_query(ip_list, 'PATCH', f'/project/{project_id}', data, token=token)

    def op_delete_project(self, ip_list: list, project_id: int, token: str = None) -> (bool, dict):
        return self.__op_query(ip_list, 'DELETE', f'/project/{project_id}', token=token)

    def op_add_code(self, ip_list: list, data: dict, token: str = None) -> (bool, dict):
        # code_id, content
        return self.__op_query(ip_list, 'POST', '/code', data, token=token)

    def op_update_code(self, ip_list: list, code_id: int, data: dict, token: str = None) -> (bool, dict):
        return self.__op_query(ip_list, 'PATCH', f'/code/{code_id}', data, token=token)

    def op_delete_code(self, ip_list: list, code_id: int, token: str = None) -> (bool, dict):
        return self.__op_query(ip_list, 'DELETE', f'/code/{code_id}', token=token)

    def op_get_node_status(self, ip: int):
        url = self.base_url.format(ip, f'/node')
        req = self.request(url, method='GET', token='operation-token')
        data = req.json()
        if data['errno'] == 0:
            return data['data']
        raise RemoteOPError('get worker error %s' % (req.json()['msg']))


class MasterMixIn(RemoteMixIn):
    """
    agent -> master 主节点向子节点进行通信
    """

    base_url = 'http://{ip}:{port}/node'.format(**FrameSettings()['MASTER'])

    def op_add_node(self, data: dict, token: str = None) -> dict:
        data = self.request(
            self.base_url,
            method='POST',
            data=data,
            token=token if token else make_token(0, 'agent')
        ).json()
        if data['errno'] == 0:
            return data['data']

        raise RemoteOPError('Call Master Failed to reg node msg %s' % (data['msg']))


class RabbitMQMixIn(object):
    authorization = HTTPBasicAuth(FrameSettings()['RABBITMQ_MANAGEMENT_CONFIG']['username'],
                                  FrameSettings()['RABBITMQ_MANAGEMENT_CONFIG']['password'])
    virtual_host = FrameSettings()['RABBITMQ_CONFIG']['virtual_host']
    base_url = '{}/%s'.format(FrameSettings()['RABBITMQ_MANAGEMENT_CONFIG']['address'])

    def request(self, url, method, params=None, data=None):
        headers = {'Content-Type': 'application/json'}
        if data:
            data = json.dumps(data)
        try:
            req = requests.request(method, url, headers=headers, data=data, params=params, auth=self.authorization)
        except ConnectionError as e:
            raise RemoteOPError(f'Call RabbitMQ Management plug-in failed please check!, error:%s' % (e))
        return req

    def op_get_project_queue_detail(self, project_id: int):
        """根据project_id 获取队列的信息"""
        uri = f'api/queues/{self.virtual_host}'
        req = self.request(self.base_url % uri, 'GET', params={'name': f'_{project_id}'})
        if req.status_code == 200:
            return req.json()
        raise RemoteOPError('get queue info Failed %s' % (req.text))

    def op_purge_queue_msg(self, project_id: int):
        uri = f'api/queues/bspider/candidate_{project_id}/contents'
        req = self.request(self.base_url % uri, 'DELETE', params={"vhost": self.virtual_host,"name":f"candidate_{project_id}","mode":"purge"})
        if req.status_code == 204:
            return None
        raise RemoteOPError('purge queue info Failed %s' % (req.text))





import hashlib
import json


class Sign(object):

    def __init__(self, **kwargs):
        plaintext = json.dumps(kwargs).encode('utf-8')
        self.__sign = hashlib.md5(plaintext).hexdigest()

    def __eq__(self, other):
        return self.__sign == str(other)

    def __str__(self):
        return self.__sign



from .exception import APIException


class Success(APIException):
    code = 200
    msg = 'ok!'
    errno = 0


class GetSuccess(Success):
    pass


class PostSuccess(Success):
    code = 201


class PutSuccess(Success):
    pass


class PatchSuccess(Success):
    pass


class DeleteSuccess(Success):
    code = 204

class PartialSuccess(Success):
    code = 206



"""
1. 初始化config文件
2. 初始化cache文件夹
3. 初始化模板
"""
import os
import string
from os import mkdir
from os.path import join, abspath
from shutil import copy2

from bspider.commands import BSpiderCommand
from bspider.utils.conf import PLATFORM_NAME_ENV, PLATFORM_PATH_ENV, BSPIDER_CONFIG_PATH_ENV
from bspider.utils.exceptions import UsageError
from bspider.utils.template import render_templatefile

TEMPLATES_TO_RENDER = (
    ('config', 'frame_settings.py.tmpl'),
    ('config', '__init__.py.tmpl'),
    ('bin', 'agent_manager.py.tmpl'),
    ('bin', 'bcorn_manager.py.tmpl'),
    ('bin', 'scheduler_manager.py.tmpl'),
    ('bin', 'master_manager.py.tmpl'),
    ('bin', '__init__.py.tmpl'),
    ('middleware', '__init__.py.tmpl'),
    ('pipeline', '__init__.py.tmpl'),
    ('projects', '__init__.py.tmpl'),
    ('.', '__init__.py.tmpl'),
)


class Command(BSpiderCommand):

    def syntax(self):
        return "<platform_name>"

    def short_desc(self):
        return "Create new platform in current path"

    def run(self, args, opts):
        if len(args) not in (1, 2):
            raise UsageError('args error')

        platform_name = args[0]

        if os.environ.get(PLATFORM_NAME_ENV):
            self.exitcode = 1
            print(f'Error: platform:{os.environ[PLATFORM_NAME_ENV]} already exists in {os.environ[PLATFORM_PATH_ENV]}')
            return

        if not self._is_valid_name(platform_name):
            self.exitcode = 1
            return

        # 拷贝模板
        self._copytree(join(self.templates_dir, 'platform'), abspath(platform_name))

        for paths in TEMPLATES_TO_RENDER:
            path = join(*paths)
            tplfile = join(platform_name,
                           string.Template(path).substitute(platform_name=platform_name))
            render_templatefile(tplfile, platform_name=platform_name)

        mkdir(join(abspath(platform_name), 'log'))
        mkdir(join(abspath(platform_name), 'log', 'supervisor'))
        mkdir(join(abspath(platform_name), '.cache'))

        tplfile = join(self.templates_dir, 'tools_cfg', '.bspider.platform.cfg.tmpl')
        copy2(tplfile, os.environ[BSPIDER_CONFIG_PATH_ENV])
        render_templatefile(os.environ[BSPIDER_CONFIG_PATH_ENV], platform_name=platform_name,
                            platform_path=abspath(platform_name))

        print(f"New BSpider platform '{platform_name}' init success")
        print(f"    template directory:\n    '{self.templates_dir}'")
        print(f"    platform at:\n    '{abspath(platform_name)}'")
        print("Please change frame_settings in ")
        print(f"{abspath(platform_name)}/frame_settings.py")
        print("You can development your first spider with:")
        print(f"    cd {abspath(platform_name)}/projects")
        print("then you can init you spider code by template")
        print("     bspider mkspider <spider_name>")


if __name__ == '__main__':
    Command().run(['hello_project'])



import yaml
from apscheduler.triggers.cron import CronTrigger
from voluptuous import Required, All, Length, Invalid, Schema, Range, MultipleInvalid
from wtforms import StringField, IntegerField

from bspider.core.api import BaseForm, ParamRequired
from bspider.utils.exceptions import ProjectSettingsError


def valid_middleware(middleware: list) -> list:
    return valid_module('Middleware', middleware)


def valid_pipeline(pipeline: list) -> list:
    data = valid_module('Pipeline', pipeline)
    pipe_count = 0
    for pipe in pipeline:
        pipeline_name = pipe
        if isinstance(pipe, dict):
            for key, _ in pipe.items():
                pipeline_name = key
        if pipeline_name.endswith('Extractor'):
            pipe_count += 1
    if pipe_count != 1:
        raise Invalid('Module Pipeline Invalid, must hive one \'Extractor\' but find %s' % (pipe_count))
    return data


def valid_module(module_type, data):
    for index, module in enumerate(data):
        if isinstance(module, str):
            data[index] = {module: dict()}
        elif isinstance(module, dict):
            for cls, param in module.items():
                if not isinstance(cls, str):
                    raise Invalid('%s cls %s name Invalid, must be str' % (module_type, cls))
                if not isinstance(param, dict):
                    raise Invalid('%s %s param Invalid, must be dict or None' % (module_type, cls))
        else:
            raise Invalid('Module %s Invalid, must like \'cls\' or [{cls: dict(cls_param)}]' % (module_type))
    return data


def valid_crontab(crontab) -> str:
    CronTrigger.from_crontab(crontab)
    return crontab


schema = Schema({
    Required('project_name'): All(str, Length(min=1, max=99)),
    Required('group', default='default'): All(str, Length(min=1, max=99)),
    Required('description', default='default'): All(str),
    Required('global_settings', default=dict()): All(dict),
    Required('downloader', default=dict()): {
        Required('max_retry_times', default=3): All(int, Range(min=1, max=10)),
        Required('ignore_retry_http_code', default=list()): All(list),
        Required('middleware', default=list()): valid_middleware
    },
    Required('parser', default=dict()): {
        Required('pipeline', default=list()): valid_pipeline
    },
    Required('scheduler'): {
        Required('rate'): All(int, Range(min=5)),
        Required('trigger'): valid_crontab,
        Required('trigger_type', default='corn'): All(str),
        Required('description', default='default'): All(str)
    }

})


class AddForm(BaseForm):
    status = IntegerField(default=1)
    config = StringField(validators=[ParamRequired()])
    editor = StringField(validators=[ParamRequired()])

    def validate_config(self, value):
        try:
            value.data = (schema(yaml.safe_load(value.data)), value.data)
        except MultipleInvalid as e:
            raise ProjectSettingsError(e.error_message)


class UpdateForm(BaseForm):
    status = IntegerField()
    config = StringField()
    editor = StringField()

    def validate_config(self, value):
        if value.data:
            try:
                value.data = (schema(yaml.safe_load(value.data)), value.data)
            except MultipleInvalid as e:
                raise ProjectSettingsError(e.error_message)



from bspider.core.api import BaseService, Conflict, NotFound, PostSuccess, DeleteSuccess, PatchSuccess, GetSuccess, \
    ParameterException
from bspider.core.api import AgentMixIn
from bspider.utils.exceptions import RemoteOPError
from bspider.master.server import log
from bspider.master.service.impl.node_impl import NodeImpl


class Node(BaseService, AgentMixIn):

    def __init__(self):
        self.impl = NodeImpl()

    # node API
    def add_node(self, ip, description, name, cpu_num, mem_size, disk_size, port):
        self.impl.add_node({
            'ip': ip,
            'description': description,
            'name': name,
            'cpu_num': cpu_num,
            'mem_size': mem_size,
            'disk_size': disk_size,
            'status': 1,
            'port': port
        })
        log.info(f'add agent from {name}->{ip}:{port} success')
        return PostSuccess(
            msg=f'add agent from {name}->{ip}:{port} success',
            data={
                'projects': self.impl.get_all_projects(),
                'codes': self.impl.get_all_codes(),
                'workers': self.impl.get_all_workers(ip)
            })

    def delete_node(self, node_id):
        """supervisor 停止进程 删除节点信息"""
        nodes = self.impl.get_node(node_id)
        if not len(nodes):
            log.error(f'node is not exist')
            return Conflict(msg=f'node is not exist', errno=20008)
        node = nodes[0]
        try:
            # 整个删除操作是一个事务
            with self.impl.mysql_client.session() as session:
                session.delete(*self.impl.delete_node(node_id, get_sql=True))
                session.delete(*self.impl.delete_worker_by_ip(node['ip'], get_sql=True))
                self.op_stop_node(node['ip'])
            log.info(f'delete node success')
            return DeleteSuccess()
        except Exception as e:
            log.error('delete node:{} failed {}'.format(node['name'], e))
            return Conflict(msg='delete node:{} failed {}'.format(node['name'], e), errno=20005)

    def update_node(self, node_id, **kwargs):
        nodes = self.impl.get_node(node_id)
        if not len(nodes):
            log.error(f'node is not exist')
            return Conflict(msg=f'node is not exist', errno=20008)
        node = nodes[0]
        try:
            remote_change = False
            for key in ('status',):
                if key in kwargs:
                    remote_change = True
                    break

            if remote_change:
                with self.impl.mysql_client.session() as session:
                    if len(kwargs):
                        session.update(*self.impl.update_node(node_id, kwargs, get_sql=True))
                    if kwargs['status'] == 1:
                        self.op_start_node(node['ip'])
                    elif kwargs['status'] == 0:
                        self.op_stop_node(node['ip'])

            elif len(kwargs):
                self.impl.update_node(node_id, kwargs)
            return PatchSuccess(msg='update node:{name} success'.format(**node))

        except Exception as e:
            log.error('update node:{} failed {} {}'.format(node['ip'], e, kwargs))
            return Conflict(msg='update node:{} failed {}'.format(node['ip'], e), errno=20005)

    def get_node(self, node_id):
        """得到node列表，查询rpc获取supervisor进程状态，整合，返回"""
        infos = self.impl.get_node(node_id)
        if infos:
            for info in infos:
                self.datetime_to_str(info)
                supervisor_node_info = self.op_get_node(info['ip'])
                info['description'] = '{}->{}'.format(info.pop('description'), supervisor_node_info.pop('description'))
                info.update(supervisor_node_info)
            return GetSuccess(msg='get node:{} status success', data=infos[0])
        log.error(f'this node id={node_id} is not exist')
        return NotFound(msg=f'this node id={node_id} is not exist', errno=20008)

    def get_nodes(self, page, limit, search, sort):
        if sort.upper() not in ['ASC', 'DESC']:
            return ParameterException(msg='sort must `asc` or `desc`')

        infos, total = self.impl.get_nodes(page, limit, search, sort)

        for info in infos:
            self.datetime_to_str(info)
            supervisor_node_info = self.op_get_node(info['ip'])
            info['description'] = '{}->{}'.format(info.pop('description'), supervisor_node_info.pop('description'))
            info.update(supervisor_node_info)

        return GetSuccess(
            msg='get nodes list success!',
            data={
                'items': infos,
                'total': total,
                'page': page,
                'limit': limit
            })

    # worker API
    def add_worker(self, ip, name, worker_type, description, status):
        """携程数量更新很麻烦所以这里写死"""
        try:
            with self.impl.mysql_client.session() as session:
                coroutine_num = 50 if worker_type == 'downloader' else 4
                data = {
                    'ip': ip,
                    'name': name,
                    'type': worker_type,
                    'description': description,
                    'coroutine_num': coroutine_num,
                    'status': status
                }
                worker_id = session.insert(*self.impl.add_worker(data, get_sql=True), lastrowid=True)
                if status == 1:
                    data = self.op_start_worker(ip, worker_id, worker_type, coroutine_num)
                else:
                    data = {'pid': 0, 'status': 0}

            log.info(f'add a new worker:{name} success')
            return PostSuccess(msg='add a new worker success!', data=data)
        except RemoteOPError as e:
            log.error(f'add a new worker:{name} failed: {e}')
            return Conflict(msg=f'add a new worker failed: {e}', errno=20003)

    def update_worker(self, worker_id, **kwargs):
        """
        name, ip
        :param worker_id:
        :param kwargs:
        :return:
        """
        workers = self.impl.get_worker(worker_id)
        if not len(workers):
            log.error(f'worker is not exist')
            return Conflict(msg=f'worker is not exist', errno=20004)
        worker = workers[0]

        try:
            remote_change = False
            for key in ('ip', 'type', 'status', 'coroutine_num'):
                if key in kwargs:
                    remote_change = True
                    break

            if remote_change:
                with self.impl.mysql_client.session() as session:
                    session.update(*self.impl.update_worker(worker_id, kwargs, get_sql=True))
                    if worker['status'] == 1:
                        self.op_stop_worker(worker['ip'], worker_id)
                        if kwargs.get('status') == 1:
                            self.op_start_worker(
                                ip=kwargs.get('ip', worker['ip']),
                                worker_type=kwargs.get('type', worker['type']),
                                worker_id=worker_id,
                                coroutine_num=kwargs.get('coroutine_num', worker['coroutine_num'])
                            )
                    elif worker['status'] == 0 and kwargs.get('status') == 1:
                        self.op_start_worker(
                            ip=kwargs.get('ip', worker['ip']),
                            worker_type=kwargs.get('type', worker['type']),
                            worker_id=worker_id,
                            coroutine_num=kwargs.get('coroutine_num', worker['coroutine_num'])
                        )

            elif len(kwargs):
                self.impl.update_worker(worker_id, kwargs)
            return PatchSuccess(msg='update worker:{name} success'.format(**worker))

        except RemoteOPError as e:
            log.error('update worker:{} failed {}'.format(worker['ip'], e))
            return Conflict(msg='update worker:{} failed {}'.format(worker['ip'], e), errno=20005)

    def delete_worker(self, worker_id):
        workers = self.impl.get_worker(worker_id)
        if not len(workers):
            log.error(f'worker is not exist')
            return Conflict(msg=f'worker is not exist', errno=20004)
        worker = workers[0]

        try:
            # 整个删除操作是一个事务
            with self.impl.mysql_client.session() as session:
                session.delete(*self.impl.delete_worker_by_id(worker_id, get_sql=True))
                self.op_stop_worker(worker['ip'], worker_id)
            return DeleteSuccess()
        except RemoteOPError as e:
            log.error('delete worker:{} failed {}'.format(worker['ip'], e))
            return Conflict(msg='delete worker:{} failed {}'.format(worker['ip'], e), errno=20009)

    def get_worker(self, worker_id):
        infos = self.impl.get_worker(worker_id)
        if len(infos):
            for info in infos:
                self.datetime_to_str(info)
                try:
                    worker_status = self.op_get_worker(info['ip'], worker_id)
                except RemoteOPError:
                    worker_status = dict(status=0, pid=0)
                info.update(worker_status)
            return GetSuccess(msg=f'get worker status success', data=infos[0])
        return NotFound(msg=f'this worker id={worker_id} is not exist', errno=20008)

    def get_workers(self, page, limit, search, sort):
        if sort.upper() not in ['ASC', 'DESC']:
            return ParameterException(msg='sort must `asc` or `desc`')

        infos, total = self.impl.get_workers(page, limit, search, sort)
        for info in infos:
            self.datetime_to_str(info)
            try:
                worker_status = self.op_get_worker(info['ip'], info['id'])
            except RemoteOPError:
                worker_status = dict(status=0, pid=0)
            info.update(worker_status)
        return GetSuccess(
            msg='get worker status success!',
            data={
                'items': infos,
                'total': total,
                'page': page,
                'limit': limit
            })



from wtforms import StringField, IntegerField

from bspider.core.api import BaseForm


class PageForm(BaseForm):
    """分页参数"""
    page = IntegerField(default=1)
    limit = IntegerField(default=8)
    search = StringField(default='')
    sort = StringField(default='DESC')

    def validate_page(self, value):
        self.__to_int(value)

    def validate_limit(self, value):
        self.__to_int(value)

    def __to_int(self, value):
        try:
            value.data = int(value.data)
        except ValueError as e:
            raise e




class ProjectSettingsError(Exception):
    pass


class ModuleError(Exception):
    pass


class ModuleExistError(ModuleError):
    pass


class DownloaderError(Exception):
    pass


class ParserError(Exception):
    pass


class RemoteOPError(Exception):
    pass


class ExtractorCallbackError(AttributeError):
    pass


class UsageError(Exception):
    pass


class MiddleWareParamError(ModuleError):
    pass


class PipelineParamError(ModuleError):
    pass


class OperationParamError(ModuleError):
    pass


class TaskParamError(ModuleError):
    pass



"""给定str 代码，加载成为对象"""
import importlib.machinery
import importlib.util
from importlib import import_module
from pkgutil import iter_modules


def import_module_by_code(module_name: str, code: str):
    """
    通过str加载第三方模块
    :param class_name:
    :param code:
    :param project_name:
    :return:
    """
    spec = importlib.util.spec_from_loader(module_name, None)
    mod = importlib.util.module_from_spec(spec)
    # 模拟exec_module()方法
    compiled_code = compile(code, f'<{module_name}>', 'exec')
    exec(compiled_code, mod.__dict__)
    return mod


def import_module_by_path(module_name: str, path: str):
    spec = importlib.util.spec_from_file_location(module_name, path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod


def walk_modules(path):
    """
    参照scrapy的方法，递归搜索整个包下面的所有模块
    """
    mods = list()
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods



from bspider.core.api import BaseService, GetSuccess
from .impl.chart_impl import ChartImpl


class ChartService(BaseService):

    def __init__(self):
        self.impl = ChartImpl()

    def parser_pv(self, project_id: int = None):
        p_total = list()
        p_error = list()
        legend = ['P.ERROR', 'P.TOTAL']

        parser_metadata = self.impl.get_project_parser_pv(project_id)

        x_axis = list()
        for metadata in parser_metadata:
            p_total.append(metadata['total'])
            p_error.append(metadata['exception'])
            x_axis.append(metadata['time'])

        return GetSuccess(data={
            'xAxis': x_axis,
            'legend': legend,
            'series': self.get_two_line_chart_series(legend, p_error, p_total)
        })

    def downloader_pv(self, project_id: int = None):
        d_total = list()
        d_error = list()
        legend = ['D.ERROR', 'D.TOTAL']

        downloader_metadata = self.impl.get_project_downloader_pv(project_id)

        x_axis = list()
        for metadata in downloader_metadata:
            d_total.append(metadata['total'])
            d_error.append(metadata['exception'])
            x_axis.append(metadata['time'])

        return GetSuccess(data={
            'xAxis': x_axis,
            'legend': legend,
            'series': self.get_two_line_chart_series(legend, d_error, d_total)
        })

    def get_two_line_chart_series(self, legend, error, total):
        return [{
            'name': legend[0],
            'itemStyle': {
                'normal': {
                    'color': '#FF005A',
                    'lineStyle': {
                        'color': '#FF005A',
                        'width': 2
                    }
                }
            },
            'smooth': True,
            'type': 'line',
            'data': error,
            'animationDuration': 2800,
            'animationEasing': 'cubicInOut'
        },
            {
                'name': legend[1],
                'smooth': True,
                'type': 'line',
                'itemStyle': {
                    'normal': {
                        'color': '#3888fa',
                        'lineStyle': {
                            'color': '#3888fa',
                            'width': 2
                        },
                        'areaStyle': {
                            'color': '#f3f8ff'
                        }
                    }
                },
                'data': total,
                'animationDuration': 2800,
                'animationEasing': 'quadraticOut'
            }]

    def get_three_line_chart_series(self, legend, cpu, memory, disk):
        return [
            {
                'name': legend[0],
                'itemStyle': {
                    'normal': {
                        'color': '#5793f3',
                        'lineStyle': {
                            'color': '#5793f3',
                            'width': 2
                        }
                    }
                },
                'smooth': True,
                'type': 'line',
                'data': cpu,
            },
            {
                'name': legend[1],
                'itemStyle': {
                    'normal': {
                        'color': '#d14a61',
                        'lineStyle': {
                            'color': '#d14a61',
                            'width': 2
                        }
                    }
                },
                'smooth': True,
                'type': 'line',
                'data': memory,
            },
            {
                'name': legend[2],
                'itemStyle': {
                    'normal': {
                        'color': '#FF005A',
                        'lineStyle': {
                            'color': '#675bba',
                            'width': 2
                        }
                    }
                },
                'smooth': True,
                'type': 'line',
                'data': disk,
            }
        ]

    def get_code_type_detail(self):
        infos = self.impl.get_code_type_detail()
        legend = list()
        for info in infos:
            legend.append(info['name'])

        return GetSuccess(data={
            'data': infos,
            'legend': legend
        })

    def node_pv(self, node_ip: str):
        cpu = list()
        memory = list()
        disk = list()
        legend = ['CPU', 'MEMORY', 'DISK']

        node_metadata = self.impl.get_node_pv(node_ip)

        x_axis = list()
        for metadata in node_metadata:
            cpu.append(metadata['cpu'])
            memory.append(metadata['memory'])
            disk.append(metadata['disk'])
            x_axis.append(metadata['time'])

        return GetSuccess(data={
            'xAxis': x_axis,
            'legend': legend,
            'series': self.get_three_line_chart_series(legend, cpu, memory, disk)
        })



from enum import Enum
from wtforms import StringField

from bspider.core.api import BaseForm, ParamRequired


class GetCodeListForm(BaseForm):
    type = StringField()


class ValidateForm(BaseForm):
    data = StringField(validators=[ParamRequired()])


class CrawlDetailTagEnum(Enum):
    sign = 'sign'
    url = 'url'


class CrawlDetailSourceEnum(Enum):
    parser = 'parser'
    downloader = 'downloader'


class GetCrawlDetailForm(BaseForm):
    tag = StringField(validators=[ParamRequired()])
    data = StringField(validators=[ParamRequired()])
    source = StringField(validators=[ParamRequired()])

    def validate_tag(self, value):
        try:
            value.data = CrawlDetailTagEnum(value.data).name
        except ValueError as e:
            raise e

    def validate_source(self, value):
        try:
            value.data = CrawlDetailSourceEnum(value.data).name
        except ValueError as e:
            raise e



import os
import sys
from configparser import ConfigParser

import bspider

BSPIDER_CONFIG_PATH_ENV = 'BSPIDER_CONFIG_PATH'
BSPIDER_VERSION_ENV = 'BSPIDER_VERSION'
PLATFORM_PATH_ENV = 'PLATFORM_PATH'
PLATFORM_NAME_ENV = 'PLATFORM_NAME'
BSPIDER_FRAME_SETTINGS_MODULE_ENV = 'BSPIDER_FRAME_SETTINGS_MODULE'


def init_platform_env(func):
    """检查是否存在platform.cfg 如存在加载配置"""
    def wrap(*args, **kwargs):
        if os.environ.get(BSPIDER_CONFIG_PATH_ENV) is None:
            bspider_config_path = os.path.expanduser(os.path.join('~', '.bspider.platform.cfg'))
            os.environ[BSPIDER_CONFIG_PATH_ENV] = bspider_config_path
            os.environ[BSPIDER_VERSION_ENV] = bspider.__version__
            if os.path.exists(bspider_config_path):
                cfg = ConfigParser()
                cfg.read([os.environ[BSPIDER_CONFIG_PATH_ENV]], encoding='utf8')
                if cfg.has_option('settings', 'default'):
                    os.environ[BSPIDER_FRAME_SETTINGS_MODULE_ENV] = cfg.get('settings', 'default')
                if cfg.has_option('deploy', 'platform'):
                    os.environ[PLATFORM_NAME_ENV] = cfg.get('deploy', 'platform')
                if cfg.has_option('deploy', 'path'):
                    os.environ[PLATFORM_PATH_ENV] = cfg.get('deploy', 'path')
                    sys.path.append(cfg.get('deploy', 'path'))
        return func(*args, **kwargs)
    return wrap



import yaml
from voluptuous import Required, All, Length, Invalid, Schema, Range

from bspider.core import Project


def valid_middleware(middleware: list) -> list:
    return valid_module('Middleware', middleware)

def valid_pipeline(pipeline: list) -> list:
    return valid_module('Pipeline', pipeline)

def valid_module(module_type, data):
    for index, module in enumerate(data):
        if isinstance(module, str):
            data[index] = { module: dict() }
        elif isinstance(module, dict):
            for cls, param in module.items():
                if not isinstance(cls, str):
                    raise Invalid('%s cls %s name Invalid, must be str' % (module_type, cls))
                if not isinstance(param, dict):
                    raise Invalid('%s %s param Invalid, must be dict or None' % (module_type, cls))
        else:
            raise Invalid('Module %s Invalid, must like \'cls\' or [{cls: dict(cls_param)}]' % (module_type))
    return data

schema = Schema({
    Required('project_name'): All(str, Length(min=1, max=99)),
    Required('rate'): All(int, Range(min=5)),
    Required('group', default='default'): All(str, Length(min=1, max=99)),
    Required('description', default='default'): All(str),
    Required('global_settings', default=dict()): All(dict),
    Required('downloader'): {
        Required('max_retry_times', default=3): All(int, Range(min=1, max=10)),
        Required('ignore_retry_http_code', default=list()): All(list),
        Required('middleware', default=list()): valid_middleware
    },
    Required('parser'): {
        Required('pipeline', default=list()): valid_pipeline
    }
})

with open('/Users/baishanglin/PycharmProjects/bspider/test/settings.yaml') as f:

    print(Project(schema(f.read())).global_settings)


class Item(object):
    """
    item对象,
    """

    def __init__(self, **kwargs):
        """
        构造函数
        """
        self.capacity = kwargs

    # 增加元素
    def __setitem__(self, key, value):
        self.capacity[key] = value

    # 取出元素
    def __getitem__(self, item):
        return self.capacity[item]

    # 删除元素
    def __delitem__(self, key):
        del self.capacity[key]

    # 对象可迭代
    def __iter__(self):
        return self.capacity.items()

    def pop(self, key):
        return self.capacity.pop(key)

    def get(self, key):
        return self.capacity.get(key)

    def keys(self):
        return self.capacity.keys()

    def values(self):
        return self.capacity.values()

    def clear(self):
        return self.capacity.clear()

    def __repr__(self):
        return f'<Item: {len(self.capacity)} field>'

    __str__ = __repr__


class MySQLSaverItem(Item):

    def __init__(self, table: str, database: str, auto_update: bool = True, **kwargs):
        """
        准备存入MySQL 的Item
        :param table: 存入mysql 的表名
        :param auto_update: 如果为true, 当插入数据已经存在会自动更新已经存在的数据
        :param kwargs: 要插入table的数据
        """
        self.table = table
        self.database = database
        self.auto_update = auto_update
        super().__init__(**kwargs)

    def __repr__(self):
        return f'<MySQLSaverItem: {self.database}->{self.table} {len(self.capacity)} field>'


class RedisSaverItem(Item):

    def __init__(self, db: int = 0, **kwargs):
        """
        Beta Item
        """
        self.db = db
        super().__init__(**kwargs)

    def __repr__(self):
        return f'<RedisSaverItem: {self.db} {len(self.capacity)} field>'


class RabbitMQSaverItem(Item):

    def __init__(self, exchange: str, **kwargs):
        """
        Beta Item
        """
        self.exchange = exchange
        super().__init__(**kwargs)

    def __repr__(self):
        return f'<RabbitMQSaverItem: {self.exchange} {len(self.capacity)} field>'



"""
** 检查是否有supervisor.pid文件确认supervisor是否启动，启动则调用rpc接口启动进程 否则：初始化supervisor.conf 启动进程
1. web service
2. 调度器
3. 定时任务管理

# 如果MySQL表未创建的实例化表和初始数据
"""
# 忽略PyMySQL警告
from warnings import filterwarnings
import pymysql
filterwarnings('ignore', category = pymysql.Warning)

import os
import re
import string
import time
from shutil import ignore_patterns, copy2

from werkzeug.security import generate_password_hash

from bspider.commands import BSpiderCommand
from bspider.utils.conf import PLATFORM_PATH_ENV, PLATFORM_NAME_ENV
from bspider.utils.database import prepare_insert_sql
from bspider.utils.exceptions import UsageError
from bspider.utils.template import render_templatefile

IGNORE = ignore_patterns('*.pyc', '.svn')


def init_custom_code(templates_dir, table):
    sql = list()
    desc = re.compile('@([A-Za-z]+)=(.*?)\n')

    for dirpath, _, filenames in os.walk(os.path.join(templates_dir, 'inner_module')):
        for filename in filenames:
            if filename.startswith('__'):
                continue

            if filename.endswith('.module'):
                with open(os.path.join(dirpath, filename)) as f:
                    text = f.read()
                    result_dict = dict(content=text)
                    param = desc.findall(text)
                    for k, v in param:
                        result_dict[k] = v.strip()
            sql.append(prepare_insert_sql(table, data=result_dict))
    return sql


PLAIN_TABLE = [
    ('bspider_cron', dict()),
    ('bspider_customcode', dict()),
    ('bspider_downloader_status', dict()),
    ('bspider_node', dict()),
    ('bspider_node_status', dict()),
    ('bspider_parser_status', dict()),
    ('bspider_project', dict()),
    ('bspider_project_customcode', dict()),
    ('bspider_user', dict(password=generate_password_hash('admin'))),
    ('bspider_worker', dict())
]


class Command(BSpiderCommand):

    def syntax(self):
        return "<op:start|stop>"

    def short_desc(self):
        return "Run/stop BSpider as a master node"

    def long_desc(self):
        return """Run/stop BSpider by supervisor with three process:
        master: a web server to manager spiders (by gunicorn and gevent).
        bcorn: a cronjob process to manager cron task.
        scheduler: dispatch all spider project.
        ** first start master service this cmd will try to create some MySQL table"""

    def init_supervisor(self):
        """查看supervisor是否已经启动"""
        platform_path = os.environ[PLATFORM_PATH_ENV]
        platform_name = os.environ[PLATFORM_NAME_ENV]

        tplfile = os.path.join(self.templates_dir, 'tools_cfg', 'master_gunicorn.py.tmpl')
        copy2(tplfile, os.path.join(platform_path, '.cache', 'master_gunicorn.py.tmpl'))
        render_templatefile(os.path.join(platform_path, '.cache', 'master_gunicorn.py.tmpl'),
                            master_port=self.frame_settings['MASTER']['port'],
                            master_ip=self.frame_settings['MASTER']['ip'],
                            log_level=self.frame_settings['LOGGER_LEVEL'].lower(),
                            platform_name=platform_name,
                            platform_path=platform_path)

        if os.path.exists(os.path.join(platform_path, '.cache', 'supervisord.pid')):
            return True
        tplfile = os.path.join(self.templates_dir, 'tools_cfg', 'supervisor.conf.tmpl')
        config_path = os.path.join(platform_path, '.cache', 'supervisor.conf')
        copy2(tplfile, config_path)
        render_templatefile(config_path,
                            platform_path=platform_path,
                            bin_path=os.path.join(platform_path, 'bin'),
                            master_ip=self.frame_settings['MASTER']['ip'],
                            supervisor_rpc_port=self.frame_settings['SUPERVISOR_RPC']['port'],
                            supervisor_rpc_username=self.frame_settings['SUPERVISOR_RPC']['username'],
                            supervisor_rpc_password=self.frame_settings['SUPERVISOR_RPC']['password'])

        cmd = 'supervisord -c {}'.format(os.path.join(platform_path, '.cache', 'supervisor.conf'))
        print('start supervisor')
        print(f'cmd: {cmd}')
        print(os.popen(cmd).read().strip())
        time.sleep(3)
        return True

    def init_database(self):
        from bspider.utils.database import MysqlClient
        mysql_client = MysqlClient.from_settings(self.frame_settings['WEB_STUDIO_DB'])

        sql = 'show tables;'
        remote_table = set()
        for table in mysql_client.select(sql):
            for table_name in table.values():
                if table_name.startswith('bspider_'):
                    remote_table.add(table_name)

        if set([table[0] for table in PLAIN_TABLE]) == remote_table:
            return True
        else:
            print('Warning: Mysql Table is not create or destroyed')
            while True:
                in_content = input("Initialize MySQL table or not (Y/N)：")
                if in_content.upper() == "N":
                    exit(1)
                elif in_content.upper() == "Y":
                    break

            # 初始化表
            for table, param in PLAIN_TABLE:
                sql_path = os.path.join(self.templates_dir, 'table_sql', f'{table}.sql')
                with open(sql_path) as f:
                    sql_list = string.Template(f.read().strip()).substitute(**param)
                    for sql in sql_list.split(';\n'):
                        if len(sql):
                            mysql_client.query(sql)
                print(f'init table:{table} success')

            sql_list = init_custom_code(self.templates_dir, 'bspider_customcode')
            print(f'insert {len(sql_list)} inner module')
            for sql, value in sql_list:
                mysql_client.insert(sql, value)

        return True

    def run(self, args, opts):
        if len(args) != 1:
            raise UsageError('args error')
        op = args[0]
        if op not in ('start', 'stop'):
            raise UsageError('unknow op: %s' % (op))

        if op == 'start':
            self.init_database()
            self.init_supervisor()
        rpc_socket = os.path.join(os.environ[PLATFORM_PATH_ENV], '.cache', 'supervisor.conf')

        print('=======supervisor output ========')
        for module in ('master', 'bcorn', 'scheduler'):
            cmd = f'supervisorctl -c {rpc_socket} {op} {module}'
            print(os.popen(cmd).read().strip())
        print('=================================')
        print(f'A new BSpider master node {op}!')
        print(f'see /platform/logs/supervisor/{module}.log to check process status!')
